#/*
# * Copyright (c) 2022 AMD Inc. All rights reserved.
# *
# * Author:
# *       Bruce Ashfield <bruce.ashfield@amd.com>
# *
# * SPDX-License-Identifier: BSD-3-Clause
# */

import sys
import types
import os
import getopt
import re
import logging
from pathlib import Path
from pathlib import PurePath
from lopper import Lopper
from lopper import LopperFmt
from lopper.tree import LopperAction
from lopper.tree import LopperTree
from lopper.tree import LopperNode
from lopper.tree import LopperProp
import lopper
import lopper.log

lopper.log._init(__name__)

def is_compat( node, compat_string_to_test ):
    if re.search( "module,extract$", compat_string_to_test):
        return extract
    return ""

def usage():
    print( """
   Usage: extract -t <target node> [OPTION]

      -t       target node (full path)
      -i       include node if found in extracted node paths
      -p       permissive matching on target node (regex)
      -v       enable verbose debug/processing
      -x       exclude nodes or properties matching regex
      -o       output file for extracted device tree

    """)

def extract( tgt_node, sdt, options ):
    try:
        verbose = options['verbose']
    except:
        verbose = 0

    try:
        args = options['args']
    except:
        args = []

    opts,args2 = getopt.getopt( args, "i:pvt:o:x:", [ "verbose", "permissive" ] )

    if opts == [] and args2 == []:
        usage()
        sys.exit(1)

    permissive=False
    exclude_list=[]
    include_list=[]
    output=None
    target_node_name=None
    for o,a in opts:
        # print( "o: %s a: %s" % (o,a))
        if o in ('-x'):
            exclude_list.append( a )
        if o in ('-i'):
            include_list.append( a )
        elif o in ('-o'):
            output=a
        elif o in ('-t'):
            target_node_name=a
        elif o in ('-v', "--verbose"):
            verbose = verbose + 1
        elif o in ('-p', "--permissive"):
            permissive = True

    if verbose > 3:
        desired_level = lopper.log.TRACE2
    elif verbose > 2:
        desired_level = lopper.log.TRACE
    elif verbose > 1:
        desired_level = logging.DEBUG
    elif verbose > 0:
        desired_level = logging.INFO
    else:
        desired_level = None

    if desired_level is not None:
        lopper.log._level(desired_level, __name__)

    lopper.log._info( f"cb: extract( {tgt_node}, {sdt}, {verbose}, {args} )" )

    if not target_node_name:
        lopper.log._error( "no target node provided" )
        usage()
        sys.exit(1)

    tgt_nodes = []
    try:
        tgt_nodes.append(sdt.tree[target_node_name])
    except:
        if permissive:
            tgt_nodes = sdt.tree.nodes( target_node_name )

    if not tgt_nodes:
        lopper.log._error( f"node {target_node_name} not found in tree" )
        if verbose > 0 or lopper.log._is_enabled(logging.INFO):
            sdt.tree.print()
        sys.exit(1)

    lopper.log._info( f"target node(s) {[y.abs_path for y in tgt_nodes]} found" )

    extracted_tree = LopperTree()

    extracted_tree_root = extracted_tree["/"]
    extracted_tree_root["#address-cells"] = 2
    extracted_tree_root["#size-cells"] = 2

    ##
    ##    - pass "interrupt-parent" as a node/property to ignore on the
    ##      command line, since we don't want it in xen
    ##
    extracted_container_node = LopperNode( -1, "/extracted" )
    extracted_container_node["compatible"] = "simple-bus"
    extracted_container_node["ranges"] = None;
    extracted_container_node["#address-cells"] = 2
    extracted_container_node["#size-cells"] = 2

    extracted_tree = extracted_tree + extracted_container_node

    for n in tgt_nodes:
        node_refs = n.resolve_all_refs( parents=False )
        refd_paths = [y.abs_path for y in node_refs]
        lopper.log._info( f"Extracted node {n.abs_path} refs: {refd_paths}" )

        for r in node_refs:
            # Don't add the parent, just ourself + phandle refs
            if r == n.parent:
                lopper.log._info( f"skipping parent of target node: {r}" )
                pass
            else:
                copy_node = True
                for exclude in exclude_list:
                    if re.search( exclude, r.name ):
                        lopper.log._info( f"skipping node (matched exclude): {r.abs_path}" )
                        copy_node = False

                if copy_node:
                    # TODO: this should look all the way up, for for now, we are
                    #       just looking one path up for the include check
                    if include_list:
                        matching_parent = None
                        if r.parent.abs_path != "/":
                            #print( "checking include list" )
                            for i in include_list:
                                if re.search( i, r.parent.name ):
                                    matching_parent = r.parent

                        if matching_parent:
                            extracted_node_copy = matching_parent()
                            for subn in extracted_node_copy.subnodes( children_only = True ):
                                if subn.name != r.name:
                                    # print( "tring to delete %s" % subn.name )
                                    extracted_node_copy - subn
                        else:
                            extracted_node_copy = r()
                    else:
                        lopper.log._info( f"adding {r}: to /extracted" )
                        extracted_node_copy = r()
                    extracted_container_node + extracted_node_copy
                    extracted_node_copy["extracted,path"] = r.abs_path

                # TODO: you are here. We also need to check the exclude list for
                #       properties, and remove them from the copied node. For now
                #       we have the same flag (-x) for nodes and properites, and may
                #       need to split it, or say that a node must have "/.*<name>"
                #       which is probably better than the split argument
                for p in extracted_node_copy:
                    for exclude in exclude_list:
                        if re.search( exclude, p.name ):
                            lopper.log._info( f"dropping masked property {p.name}" )
                            extracted_node_copy - p


    extracted_tree.strict = False
    extracted_tree.resolve()

    sdt.subtrees["extracted"] = extracted_tree

    if output:
        sdt.write( extracted_tree, output, True, True )

    return True



#/*
# * Copyright (c) 2019,2020 Xilinx Inc. All rights reserved.
# *
# * Author:
# *       Bruce Ashfield <bruce.ashfield@xilinx.com>
# *
# * SPDX-License-Identifier: BSD-3-Clause
# */

import argparse
import copy
import struct
import sys
import types
import unittest
import os
import getopt
import re
import subprocess
import shutil
from pathlib import Path
from pathlib import PurePath
from io import StringIO
import contextlib
import importlib
from lopper import Lopper
from lopper import LopperFmt
import lopper
from lopper.tree import *
from re import *
from string import Template

sys.path.append(os.path.dirname(__file__))
from openamp_xlnx_common import *
from baremetalconfig_xlnx import get_cpu_node
from string import ascii_lowercase as alc

RPU_PATH = "/rpu@ff9a0000"
REMOTEPROC_D_TO_D = "openamp,remoteproc-v1"
REMOTEPROC_D_TO_D_v2 = "openamp,remoteproc-v2"
RPMSG_D_TO_D = "openamp,rpmsg-v1"
LIBMETAL_D_TO_D = "libmetal,ipc-v1"

def is_compat( node, compat_string_to_test ):
    """Identify whether this plugin handles the provided compatibility string.

    Args:
        node (LopperNode): Device tree node being evaluated. Present to satisfy the
            dispatcher interface; not used for the decision.
        compat_string_to_test (str): Compatibility string extracted from the node.

    Returns:
        Callable | str: ``xlnx_openamp_rpu`` when the compatibility string matches,
        otherwise an empty string to indicate no match.

    Algorithm:
        Performs a regular-expression search for ``openamp,xlnx-rpu`` within the
        provided string and returns the registered handler on success.
    """
    if re.search( "openamp,xlnx-rpu", compat_string_to_test):
        return xlnx_openamp_rpu
    return ""


def xlnx_openamp_keep_node(linux_dt, zephyr_dt, node, tree):
    """Report whether a node shode stay for OpenAMP Use cases.

    Args:
        linux_dt (bool): True if for Linux domain. Else False.
        zephyr_dt (bool): True if for Zephyr domain. Else False.
        node (LopperNode): Node to check
        tree (LopperTree): Tree for lopper nodes.
    Returns:
        True if Node should remain. Else False.

    Algorithm:
        Try each condition for the given node.
    """
    if not isinstance(node, LopperNode):
        print("OPENAMP: XLNX: ERROR: expected node ref in xlnx_openamp_keep_node")
        return False

    conditions = [
        "uio" in node.propval('compatible', list) and not zephyr_dt,
        "vnd,mbox-consumer" in node.propval('compatible', list),
        "zephyr,mbox-ipm" in node.propval('compatible', list),
    ]

    return any(c for c in conditions if c)


def xlnx_openamp_trim_timers(sdt, target_os, machine):
    """Trim tree of timers based on OS.

    Args:
        sdt(LopperTree): Tree for lopper nodes.
        target_os (str): OS for this lopper run
        machine (str): machien coresponding to the domain.
    Returns:
        True if success. Else False.

    Algorithm:
        Find domains matching target OS.
        For each relation, look for timer property.
        Based on timer property trim irrelevant timers.
    """
    tree = sdt.tree

    if get_platform(tree, 0) in [ SOC_TYPE.VERSAL2, SOC_TYPE.VERSAL_NET ]:
        return False

    # first filter by linux domain
    if target_os == "linux_dt":
        domains = [ n for n in tree["/domains"].subnodes(children_only=True) if n.propval("os,type") == [target_os.replace("_dt","")] ]
    else:
        # otherwise filter by machine
        match_cpunode = get_cpu_node(sdt, {'args':[machine]})
        domains = [ n for n in tree["/domains"].subnodes(children_only=True) if match_cpunode.parent == sdt.tree.pnode(n.parent.parent.propval("cpus")[0]) ]


    timer_pvals = [ n.propval("timer") for n in domains[0].subnodes(children_only=True) if n.propval("timer") != [''] ]

    # only do trim if timer prop is provided
    if timer_pvals == []:
        return False

    flattend_timer_pvals = [item for sublist in timer_pvals for item in sublist]
    relevant_timer_nodes = [ tree.pnode(phandle) for phandle in flattend_timer_pvals ]

    # enable UIO timers
    for i in relevant_timer_nodes:
        if "uio" in i.propval("compatible"):
            i['status'] = 'okay'

    all_timer_nodes = [ n for n in tree["/axi"].subnodes(children_only=True, name="timer@*") if n.propval("compatible") == ["cdns,ttc"] ]

    # remove UIO timers from stripping
    if os == "linux_dt":
        [ all_timer_nodes.remove(timer_node) for timer_node in all_timer_nodes if "uio" in timer_node.propval("compatible") ]

    # delete irrelevant timer nodes
    [ tree.delete(timer_node) for timer_node in all_timer_nodes if timer_node not in relevant_timer_nodes ]

    return True


def xlnx_handle_relations(sdt, machine, find_only = True, os = None):
    """Process OpenAMP relation domains for a given machine.

    Args:
        sdt (LopperSDT): Structured device tree wrapper containing the parsed tree.
        machine (str): Name of the remote machine to target.
        find_only (bool): When True, return the first matching domain without
            modifying the tree.
        os (str | None): Operating-system context (for example ``linux_dt``).

    Returns:
        LopperNode | bool | None: Matching domain when ``find_only`` is True, True
        when processing succeeds, False on failure, or None when the search mode
        finds no relations.

    Algorithm:
        Resolves the CPU node associated with the requested machine, collects
        remoteproc and RPMsg relation nodes, delegates processing to the dedicated
        parsers, tracks carveouts for later validation, and verifies the resulting
        reserved-memory layout does not contain overlaps. Emits a warning when no
        relations are discovered and exits without modifying the tree.
    """
    tree = sdt.tree

    # get_cpu_node expects dictionary where first arg first element is machine
    match_cpunode = get_cpu_node(sdt, {'args':[machine]})
    if not match_cpunode:
        print("xlnx_handle_relations: unable to find machine: ", machine)
        return False

    parse_routines = { REMOTEPROC_D_TO_D_v2: xlnx_remoteproc_parse, RPMSG_D_TO_D: xlnx_rpmsg_parse, LIBMETAL_D_TO_D : xlnx_rpmsg_parse }

    # first collect all relevant openamp domains
    remoteproc_relations = []
    rpmsg_relations = []

    for n in sdt.tree["/domains"].subnodes():
        node_compat = n.propval("compatible")[0]
        if node_compat == '':
            continue

        if n.parent.parent.propval("cpus") == ['']:
            continue

        # ensure target domain matches
        if match_cpunode.parent == sdt.tree.pnode(n.parent.parent.propval("cpus")[0]):
            if find_only:
                 return n
            else: # do processing on found ndoes
                if node_compat in parse_routines.keys():
                    arr = remoteproc_relations if node_compat == REMOTEPROC_D_TO_D_v2 else rpmsg_relations

                    # skip libmetal relations for Linux
                    if os == "linux_dt" and node_compat == LIBMETAL_D_TO_D:
                        continue

                    arr.append(n)

    # As the RPMsg relation will be appending nodes to a remoteproc node, link the rpmsg
    # relation to its corresponding remoteproc relation so the remoteproc relation can pass along
    # the remoteproc core node
    remoteproc_core_mapping_to_rpmsg_relation = {}

    # used to check if conflicts in ELFLOAD and IPC carveouts
    carveout_validation_arr = []

    for rel in remoteproc_relations:
        ret = xlnx_remoteproc_parse(tree, rel, carveout_validation_arr, 1)
        if not ret:
            return ret

        # save remoteproc core nodes for later use in rpmsg processing
        remoteproc_core_mapping_to_rpmsg_relation.update(ret)

    for rel in rpmsg_relations:
        if not xlnx_rpmsg_parse(tree, rel, machine, carveout_validation_arr, remoteproc_core_mapping_to_rpmsg_relation, os, 1):
            return False

    # check if conflicts in ELFLOAD and IPC carveouts
    if not xlnx_validate_carveouts(tree, carveout_validation_arr):
            return False

    if not remoteproc_relations and not rpmsg_relations:
        print("OPENAMP: XLNX: WARNING: no remoteproc or rpmsg relations found for machine", machine)

    # if here for find case, then return None as failure
    # if processing too and we are here, then this did not encounter error. So return True.
    # note that if the tree does not have openamp nodes True is also returned.
    return None if find_only else True

def xlnx_rpmsg_update_tree_linux(tree, node, ipi_node, core_node, rpmsg_carveouts, verbose = 0 ):
    """Inject Linux-specific RPMsg properties into the device tree.

    Args:
        tree (LopperTree): Device tree being updated.
        node (LopperNode): RPMsg relation child describing a channel endpoint.
        ipi_node (LopperNode): Interrupt node used for mailbox communication.
        core_node (LopperNode): Remoteproc core node associated with the channel.
        rpmsg_carveouts (list[LopperNode]): Reserved-memory nodes assigned to RPMsg.
        verbose (int): Verbosity flag controlling diagnostic output.

    Returns:
        bool: True when the tree is updated successfully, False on validation errors.

    Algorithm:
        Validates carveout naming, promotes the buffer carveout to the front of the
        memory-region list, appends carveouts to the core node, reorders DDR boot
        entries to trail RPMsg carveouts, and injects mailbox properties required by
        the Linux remoteproc driver.
    """
    print(" -> xlnx_rpmsg_update_tree_linux", node)
    vdev0buf_node = [n for n in rpmsg_carveouts if "vdev0buf" in n.name]
    if len(vdev0buf_node) == 1:
        vdev0buf_node[0] + LopperProp(name="compatible", value="shared-dma-pool")
    else:
        print("ERROR: missing or multiple vdev0buf nodes for linux rpmsg case")
        return False

    # vdev0buf should be first after the ELF load prop already in memory-region
    vdev0buf = [ index for index, rc in enumerate(rpmsg_carveouts) if "vdev0buffer" in rc.name ]
    if len(vdev0buf) != 1:
        print("ERROR: xlnx_rpmsg_update_tree_linux: expected 1 vdev0buffer node. got ", vdev0buf)
        return False

    vdev0buf = rpmsg_carveouts.pop(vdev0buf[0])
    rpmsg_carveouts.insert(0, vdev0buf)
    new_mem_region_prop_val = core_node.propval("memory-region")

    [ new_mem_region_prop_val.append(rc.phandle) for rc in rpmsg_carveouts ]

    # If DDRBOOT, ensure that it is after RPMSG carveouts
    # # save ddrboot node and add to end of list
    ddrboot_node_index = [ index for index, phandle in enumerate(new_mem_region_prop_val) if "ddrboot" in tree.pnode(phandle).name ]
    if ddrboot_node_index != []:
        new_mem_region_prop_val.append( new_mem_region_prop_val.pop(ddrboot_node_index[0]) )

    # update property with new values
    core_node["memory-region"].value = new_mem_region_prop_val
    core_node + LopperProp(name="mboxes", value = [ipi_node.phandle, 0, ipi_node.phandle, 1])
    core_node + LopperProp(name="mbox-names", value = ["tx", "rx"])
    return True

# Inputs: openamp-processed SDT, target processor
# If there exists a DDR carveout for ELF-Loading, return the start and size
# of the carveout
def xlnx_openamp_get_ddr_elf_load(machine, sdt):
    """Retrieve the DDR ELF-load carveout for a target machine.

    Args:
        machine (str): Identifier for the remote processor of interest.
        sdt (LopperSDT): Processed device tree structure.

    Returns:
        tuple[int, int] | bool: Tuple of (base, size) when a carveout exists, or
        False if the carveout cannot be resolved.

    Algorithm:
        Matches the machine to its CPU node, locates the associated RPMsg relation,
        validates host references, iterates remoteproc relations, and returns the
        ``reg`` data from the first ELFLOAD node mapped to DDR.
    """
    # get_cpu_node expects dictionary where first arg first element is machine
    match_cpunode = get_cpu_node(sdt, {'args':[machine]})
    if not match_cpunode:
        print("unable to find machine: ", machine)
        return False

    # map machine to CPU node and then to openamp domain with relation
    target_node = None
    for n in sdt.tree["/domains"].subnodes(children_only=True):
        node_compat = n.propval("compatible")[0]
        # domains must (a) have compatible and (b) relate to CPU
        if node_compat == '' or n.parent.parent.propval("cpus") == ['']:
            continue

        # ensure target domain matches
        if match_cpunode.parent == sdt.tree.pnode(n.parent.parent.propval("cpus")[0]):
             target_node = n
             break

    if target_node == None:
        print("OPENAMP: XLNX: ERROR: unable to map machine", machine, "to relation")
        return False

    # find node described in the domain that is for ELF LOAD

    # remote should only have one relevant host
    rpmsg_rels = target_node.subnodes(children_only=True)
    if len(rpmsg_rels) != 1:
        print("OPENAMP: XLNX: ERROR: expected 1 and only 1 rpmsg relation for ", target_node)
        return False

    rpmsg_rel = rpmsg_rels[0]
    if rpmsg_rel.propval("host") == [''] and len(rpmsg_rel.propval("host")) != 1:
        print("OPENAMP: XLNX: ERROR: expected host prop for", target_node)
        return False

    host = rpmsg_rel.propval("host")
    host_node = sdt.tree.pnode(rpmsg_rel.propval("host")[0])
    if not isinstance(host_node, LopperNode):
        print("OPENAMP: XLNX: ERROR: expected host node ref in host prop for", rpmsg_rel)
        return False

    if target_node.propval('compatible') == ["libmetal,ipc-v1"]:
        for rel in target_node.subnodes(children_only=True):
            elfload = rel.propval("elfload")
            if elfload == ['']:
                print("OPENAMP: XLNX: ERROR: libmetal remote domain needs elfload property.")
            elfload_node = sdt.tree.pnode(elfload[0])
            reg_val = elfload_node.propval("reg")
            return (reg_val[1], reg_val[3], "LIBMETAL_DDR")
        print("OPENAMP: XLNX: ERROR: libmetal invalid domain setup.")
        return False

    # look through host for matching remoteproc relation. If found then return the relation's elfload property reg value
    for rel in host_node.subnodes(children_only=True):
        if rel.propval("remote") != [''] and ['openamp,remoteproc-v2'] == rel.parent.propval("compatible"):
            if rel.propval("remote") == ['']:
                print("OPENAMP: XLNX: ERROR: elfload needs remoteproc host to describe elfload region")
                return False

            #  check that the referenced remote matches
            referenced_remote_domain = sdt.tree.pnode(rel.propval("remote")[0])
            if referenced_remote_domain == None or referenced_remote_domain != target_node.parent.parent:
                print("OPENAMP: XLNX: ERROR: referenced remote is invalid for host", rel)
                return False

            elfload_nodes = [ sdt.tree.pnode(i) for i in rel.propval("elfload") ]
            relevant_elfload_nodes = [ i for i in elfload_nodes if i != None and 'mmio-sram' not in i.propval('compatible')]
            if relevant_elfload_nodes == []:
                print("OPENAMP: XLNX: ERROR: expected at least one ELFLOAD node for case of generating openamp linker script using DDR.")
                return False

            # return reg from match
            reg_val = relevant_elfload_nodes[0].propval("reg")
            if reg_val == ['']:
                print("OPENAMP: XLNX: ERROR: expected 'reg' property for elfload entry", relevant_elfload_nodes[0])
                return False

            return (reg_val[1], reg_val[3], "RSC_TABLE")

    print("OPENAMP: XLNX: ERROR: unable to find elf load carveout")
    return False

# Inputs: openamp-processed SDT, target processor, ipi, ipc node
def xlnx_rpmsg_update_tree_zephyr(machine, tree, ipi_node, domain_node, ipc_nodes, relation_compat):
    """Tailor the device tree for Zephyr RPMsg communication.

    Args:
        machine (str): Remote machine identifier (unused, present for symmetry).
        tree (LopperTree): Device tree being updated.
        ipi_node (LopperNode): IPI node used for mailbox signaling.
        domain_node (LopperNode): Domain node. It may contain ddr boot field.
        ipc_nodes (list[LopperNode]): IPC shared-memory nodes tied to RPMsg.
        relation_compat (str): relation compatible string

    Returns:
        bool: True when Zephyr-specific updates succeed, False on validation failure.

    Algorithm:
        Validates IPC node count, sets Zephyr chosen-node properties, injects a
        mailbox consumer helper, removes alternative IPI siblings to avoid conflicts,
        and clears flash/OCM choices that would clash with RPMsg shared memory.
    """

    if len(ipc_nodes) != 3:
        print("ERROR: zephyr: rpmsg: only length of 3 ipc node allowed. got: ", ipc_nodes)
        return False

    # have to now combine the 3 IPC nodes into one.
    base = 0xFFFFFFFF
    size = 0
    for i in ipc_nodes:
        reg = i['reg']
        size += reg[3]
        if reg[1] < base:
            base = reg[1]

    # remove current IPC nodes. Create combined one below.
    [ tree - node for node in ipc_nodes ]

    new_ipc_node = LopperNode(-1, "/reserved-memory/ipc@%s" % hex(base)[2:])
    new_ipc_node + LopperProp(name="reg", value=[0, base, 0, size])
    new_ipc_node + LopperProp(name="compatible", value="mmio-sram")

    tree + new_ipc_node
    tree['/chosen']['zephyr,ipc_shm'] = new_ipc_node.abs_path

    if domain_node.propval("xlnx,ddr-boot") != []:
        elfload_nodes = [ tree.pnode(x) for x in domain_node.propval("reserved-memory") ]
        valid_elfload_node = [ node for node in elfload_nodes if node and node.propval("device_type") == ['memory'] ]
        if len(valid_elfload_node) > 0:
            tree['/chosen']['zephyr,sram'] = valid_elfload_node[0].abs_path

    # only create a node for this the first time. in the future this will go away as upstream wants use of ipm mbox node. this is here for bkwd compatibility
    try:
        # if here then mbox_consumer_node was already created.
        mbox_consumer_node = tree['/mbox-consumer']
    except KeyError:
        mbox_consumer_node = LopperNode(-1, "/mbox-consumer")
        mbox_consumer_props = { "compatible" : 'vnd,mbox-consumer', "mboxes" : [ipi_node.phandle, 0, ipi_node.phandle, 1], "mbox-names" : ['tx', 'rx'] }
        [mbox_consumer_node + LopperProp(name=n, value=mbox_consumer_props[n]) for n in mbox_consumer_props.keys()]
        tree.add(mbox_consumer_node)

    mbox_ipm_node = LopperNode(-1, "/mbox_ipi_%s_%s" % (hex(ipi_node['reg'][1])[2:], hex(ipi_node.parent['reg'][1])[2:]))
    mbox_ipm_props = { "compatible" : "zephyr,mbox-ipm", "mbox-names" : ['tx', 'rx'], "status": "okay", "mboxes" : [ipi_node.phandle, 0, ipi_node.phandle, 1] }
    [mbox_ipm_node +  LopperProp(name=n, value=mbox_ipm_props[n]) for n in mbox_ipm_props.keys()]
    tree.add(mbox_ipm_node)

    # do this for upstream compatibility for now
    if RPMSG_D_TO_D == relation_compat:
        tree['/chosen']['zephyr,ipc'] = mbox_ipm_node.abs_path

    if tree['/chosen'].propval('zephyr,flash') != ['']:
        tree['/chosen'].delete(sdt.tree['/chosen']['zephyr,flash'])
    if tree['/chosen'].propval('zephyr,ocm') != ['']:
        tree['/chosen'].delete(sdt.tree['/chosen']['zephyr,ocm'])

    # if user passes in xlnx,zephyr,mems - then update device_type for those referenced nodes.
    xlnx_zeph_mems = domain_node.propval("xlnx,zephyr,mems")
    if xlnx_zeph_mems != [''] and isinstance(xlnx_zeph_mems, list):
        for i in domain_node.propval("xlnx,zephyr,mems"):
            xlnx_zeph_mem_node = tree['/'].subnodes(children_only=True, name=domain_node.propval("xlnx,zephyr,mems")[0])
            if xlnx_zeph_mem_node != []:
                xlnx_zeph_mem_node[0]['device_type'] = "memory"
                tree['/chosen']['zephyr,sram'] = xlnx_zeph_mem_node[0].abs_path

    return True

def xlnx_libmetal_gen_output_file(tree, output_file, carveouts, ipi_node, timer_node, os, verbose = 0 ):
    """Generate .cmake file for Libmetal IPI usage

    Args:
        tree (LopperTree): Device tree being inspected for metadata.
        output_file (str): name of output file
        carveouts (list[LopperNode]): Carveouts associated with libmetal
        ipi_node (LopperNode): IPI node used
        timer_node (LopperNode): Timer node used
        os (str): os value
        verbose (int): Verbosity flag for diagnostic printing.

    Returns:
        bool: True on successful file generation, False on failure.
    """
    platform = get_platform(tree, verbose)
    if platform == None:
        return False
    desc0 = carveouts[0]
    desc1 = carveouts[1]
    data = carveouts[2]

    suffix = "ipi" if platform == SOC_TYPE.ZYNQMP else "mailbox"

    parent_ipis = tree["/axi"].subnodes(children_only=True, name="%s@*" % suffix)
    relevant_remote_ipi = [n for n in parent_ipis if n.propval("xlnx,ipi-bitmask")[0] == ipi_node.propval("xlnx,ipi-bitmask")[0] ]
    values = {  "SHM_IMAGE_BASE": hex(data["reg"][1]), "SHM_IMAGE_SIZE": hex(data["reg"][3]) }
    values.update({
                "SHM_PAYLOAD_BASE": values["SHM_IMAGE_BASE"], "SHM_PAYLOAD_SIZE": values["SHM_IMAGE_SIZE"],
                "SHM_PAYLOAD_HALF_SIZE": hex(data["reg"][3]//2),
                "SHM_PAYLOAD_RX_OFFSET": "0x0",
                "SHM_BASE_ADDR": values["SHM_IMAGE_BASE"], "SHM_SIZE": values["SHM_IMAGE_SIZE"],
                "SHM0_DESC_BASE": hex(desc0["reg"][1]), "SHM0_DESC_SIZE": hex(desc0["reg"][3]),
                "SHM1_DESC_BASE": hex(desc1["reg"][1]), "SHM1_DESC_SIZE": hex(desc1["reg"][3]),
                "TTC_DEV_NAME": "%s.timer" % hex(timer_node["reg"][1])[2:], "TTC_NODEID": hex(timer_node.propval("power-domains")[1]),
                "TTC_BASE_ADDR": hex(timer_node["reg"][1]),
                "IPI_DEV_NAME": "%s.%s" % (hex(ipi_node.parent["reg"][1])[2:], suffix), "IPI_BASE_ADDR": hex(ipi_node.parent["reg"][1]),
                "IPI_MASK": hex(ipi_node['xlnx,ipi-bitmask'].value[0]),
                "IPI_IRQ_VECT_ID": 0 if os == "linux_dt" else relevant_remote_ipi[0].propval("xlnx,int-id")[0],
                "BUS_NAME": "platform" if os == "linux_dt" else "generic" })
    values.update({"SHM_PAYLOAD_TX_OFFSET": values["SHM_PAYLOAD_HALF_SIZE"]})
    values.update({"SHM_DEV_NAME": "%s.%s" % (data.name.split("@")[1].lower(), data.name.split("@")[0].lower())})
    values.update({"SHM0_DESC_DEV_NAME": "%s.%s" % (desc0.name.split("@")[1].lower(), desc0.name.split("@")[0].lower())})
    values.update({"SHM1_DESC_DEV_NAME": "%s.%s" % (desc1.name.split("@")[1].lower(), desc1.name.split("@")[0].lower())})

    if os not in [ "linux_dt", "baremetal_dt" ]:
        print("unsupported os:", os)
        return False

    try:
        with open(output_file, "w") as f:
            output = Template(libmetal_cmake_template)
            f.write(output.substitute(values))
    except Exception as e:
        print("OPENAMP: XLNX: ERROR: xlnx_libmetal_gen_output_file: Error in generating template for RPU header.", e)
        return False

    return True

def xlnx_openamp_gen_outputs_only(tree, machine, output_file, memory_region_nodes, host_ipi, verbose = 0 ):
    """Generate C header output for OpenAMP RPMsg channels.

    Args:
        tree (LopperTree): Device tree being inspected for metadata.
        machine (str): Remote machine identifier (unused directly but retained for
            debugging).
        output_file (str): Destination filepath for the generated header.
        memory_region_nodes (list[LopperNode]): Carveouts associated with RPMsg.
        host_ipi (LopperNode): IPI node connected to the host processor.
        verbose (int): Verbosity flag for diagnostic printing.

    Returns:
        bool: True on successful file generation, False on failure.

    Algorithm:
        Aggregates VRING and buffer sizes from carveouts, extracts IPI configuration
        data, prepares a template substitution dictionary, and writes the rendered
        header to the requested output path.
    """
    vrings = [n for n in memory_region_nodes if 'vring' in n.name]
    vring_total_sz = hex(sum([n.propval("reg")[3] for n in vrings]))
    shm_pa = hex(min([n.propval("reg")[1] for n in vrings]))
    shbuf_sz = hex([n.propval("reg")[1] for n in memory_region_nodes if 'vdev0buffer' in n.name][0])

    remote_ipi = host_ipi.parent

    remote_vect_id = remote_ipi.propval('xlnx,int-id')[0]
    ipi_irq_vect_id = hex(remote_vect_id)
    ipi_irq_vect_id_rtos = hex(remote_vect_id-32)
    remote_ipi_str = hex(remote_ipi.propval('reg')[1])
    host_bitmask = hex(host_ipi.propval('xlnx,ipi-bitmask')[0])

    try:
        inputs = {
        "POLL_BASE_ADDR": remote_ipi_str,
        "SHM_DEV_NAME": "\"x.shm\"",
        "DEV_BUS_NAME": "\"generic\"",
        "IPI_DEV_NAME":  "\"y.ipi\"",
        "IPI_IRQ_VECT_ID": ipi_irq_vect_id,
        "IPI_IRQ_VECT_ID_FREERTOS": ipi_irq_vect_id_rtos,
        "IPI_CHN_BITMASK": host_bitmask,
        "RING_TX": "FW_RSC_U32_ADDR_ANY",
        "RING_RX": "FW_RSC_U32_ADDR_ANY",
        "SHARED_MEM_PA": shm_pa,
        "SHARED_MEM_SIZE": shbuf_sz,
        "SHARED_BUF_OFFSET": vring_total_sz,
        "EXTRAS":"",
        }

        with open(output_file, "w") as f:
            output = Template(platform_info_header_r5_template)
            f.write(output.substitute(inputs))
    except Exception as e:
        print("OPENAMP: XLNX: ERROR: xlnx_openamp_gen_outputs_only: Error in generating template for RPU header.", e)
        return False

    return True

def xlnx_rpmsg_parse(tree, rpmsg_relation_node, machine, carveout_validation_arr, channel_to_core_dict = None, os = None, verbose = 0 ):
    """Parse RPMsg relations and update the device tree accordingly.

    Args:
        tree (LopperTree): Device tree being modified.
        rpmsg_relation_node (LopperNode): Domain relation describing RPMsg channels.
        machine (str): Remote machine identifier (used for logging and lookups).
        carveout_validation_arr (list[LopperNode]): Accumulator for carveouts to be
            validated later.
        channel_to_core_dict (dict[str, LopperNode] | None): Mapping of remote names
            to remoteproc core nodes.
        os (str | None): Operating-system context (``linux_dt`` or ``zephyr_dt``).
        verbose (int): Verbosity flag for diagnostic messages.

    Returns:
        bool: True when parsing succeeds, False if required metadata is missing.

    Algorithm:
        Determines the platform, iterates RPMsg endpoints, validates remote/host
        references, gathers carveouts, applies OS-specific tree rewrites, and
        optionally emits a header file via ``xlnx_openamp_gen_outputs_only``.
    """
    print(" -> xlnx_rpmsg_parse", rpmsg_relation_node)

    platform = get_platform(tree, verbose)
    if platform == None:
        return False

    for node in rpmsg_relation_node.subnodes(children_only=True):
        pname = "remote" if os == "linux_dt" else "host"
        # check for remote property
        if node.props(pname) == []:
            print("ERROR: ", node, "is missing remote property")
            return False

        remote_node = tree.pnode(node.propval(pname)[0])
        if remote_node == None:
             print("ERROR: invalid rpmsg ", pname," for relation: ", rpmsg_relation_node)
             return False

        if os == "linux_dt" and remote_node.name not in channel_to_core_dict.keys():
            print("ERROR: missing needed remoteproc core node for rpmsg parse to occur.")
            return False

        core_node = channel_to_core_dict[remote_node.name] if os == "linux_dt" else None

        # first find host to remote IPI
        mbox_pval = node.propval("mbox")
        if mbox_pval == ['']:
            print("ERROR: ", node, " is missing mbox property")
            return False

        ipi_node = tree.pnode(mbox_pval[0])
        if ipi_node == None:
            print("ERROR: Unable to find ipi")
            return False

        carveouts_node = tree[node.parent.parent.parent.abs_path + "/domain-to-domain/rpmsg-relation"]
        carveout_prop = node.propval("carveouts")
        if carveout_prop == ['']:
            print("ERROR: ", node, " is missing carveouts property")
            return False

        rpmsg_carveouts = [ tree.pnode(phandle) for phandle in carveout_prop ]

        # validate later
        carveout_validation_arr.extend(rpmsg_carveouts)

        # until domain access is in place - need to manually prune some nodes
        try:
            res_mem_node = tree["/reserved-memory"]
            [ tree.delete(i) for i in res_mem_node.subnodes() if i.propval("compatible") == ['mmio-sram'] and os == "linux_dt" ]
            if res_mem_node.propval("ranges") == [1]:
                res_mem_node.delete("ranges")
                res_mem_node + LopperProp(name="ranges")
        except KeyError:
            print("ERROR: carveouts should be in reserved memory.")
            return False

        if os == "zephyr_dt" and not xlnx_rpmsg_update_tree_zephyr(machine, tree, ipi_node, node.parent.parent.parent, rpmsg_carveouts, rpmsg_relation_node.propval("compatible")[0]):
            return False
        if os == "linux_dt"  and not xlnx_rpmsg_update_tree_linux(tree, node, ipi_node, core_node, rpmsg_carveouts, verbose):
            return False

    return True

# tests for a bit that is set, going fro 31 -> 0 from MSB to LSB
def check_bit_set(n, k):
    """Check whether the k-th bit within an integer is set.

    Args:
        n (int): Value to test.
        k (int): Bit index to inspect.

    Returns:
        bool: True when the bit is set, otherwise False.

    Algorithm:
        Applies a bitmask constructed via ``1 << k`` and performs a bitwise AND,
        returning True when the result is non-zero.
    """
    if n & (1 << (k)):
        return True

    return False


def determine_cpus_config(remote_domain):
    """Map the remote domain CPU configuration string to an enum value.

    Args:
        remote_domain (LopperNode): Domain node describing the remote processor.

    Returns:
        CPU_CONFIG | int: CPU configuration enum for split/lockstep, or -1 on error.

    Algorithm:
        Validates that ``cpu_config_str`` exists, ensures the value is one of the known
        strings, and converts it into the matching ``CPU_CONFIG`` enum constant.
    """
    print(" -> determine_cpus_config ", remote_domain, remote_domain.propval("cpu_config_str"), remote_domain.propval("cpus"))
    if remote_domain.propval("cpu_config_str") == ['']:
        print(" determine_cpus_config failed. could not find cpu_config_str property on remote domain", remote_domain)
        return -1

    if remote_domain.propval("cpu_config_str") not in [ ['split'], ['lockstep'] ]:
        print(" determine_cpus_config failed. invalid cpu_config_str: ", remote_domain.propval("cpu_config_str"))
        return -1

    return { "split": CPU_CONFIG.RPU_SPLIT, "lockstep": CPU_CONFIG.RPU_LOCKSTEP }[remote_domain.propval("cpu_config_str")[0]]

def determinte_rpu_core(tree, cpu_config, remote_node):
    """Determine which RPU core index is used for the remote node.

    Args:
        tree (LopperTree): Device tree containing the remote node.
        cpu_config (CPU_CONFIG): RPU configuration (split or lockstep).
        remote_node (LopperNode): Remote node describing the remote processor.

    Returns:
        RPU_CORE | bool: Enum representing the selected core, or False on failure.

    Algorithm:
        Validates the presence of ``core_num`` and converts it into the appropriate
        ``RPU_CORE`` enum instance.
    """
    print(" -> determinte_rpu_core", cpu_config, remote_node)
    if remote_node.propval("core_num") == ['']:
        print(" determinte_rpu_core failed. could not find core_num property no node: ", remote_node)
        return False

    core_index = int(remote_node.propval("core_num")[0])
    return RPU_CORE(core_index)


def cells_to_int(cells):
    val = 0
    for c in cells:
        val = (val << 32) | c
    return val


def xlnx_validate_carveouts(tree, carveouts):
    """Verify that carveout regions do not overlap within reserved memory.

    Args:
        tree (LopperTree): Device tree containing reserved-memory nodes.
        carveouts (list[LopperNode]): Carveout nodes to validate.

    Returns:
        bool: True when no overlaps are detected and reserved-memory exists, False
        otherwise.

    Algorithm:
        Ensures the ``/reserved-memory`` node exists, gathers ``reg`` tuples from
        carveouts, and checks for pairwise overlap among relevant regions.
    """
    print(" -> xlnx_validate_carveouts")
    expect_ddr = any(["/reserved-memory/" in n.abs_path for n in carveouts])
    try:
        res_mem_node = tree["/reserved-memory"]
    except KeyError:
        if expect_ddr:
            print("ERROR: carveouts should be in reserved memory.")
            return False

        res_mem_node = LopperNode(-1, "/reserved-memory")
        res_mem_node + LopperProp(name="ranges",value=[])
        tree.add(res_mem_node)

    if res_mem_node.propval('#size-cells') == [''] or res_mem_node.propval('#address-cells') == ['']:
        print("ERROR: malformed reserved memory - expected #size-cells and #address-cells")
        return False

    addr_cells = res_mem_node.propval('#address-cells')[0]
    size_cells = res_mem_node.propval('#size-cells')[0]


    carveout_pairs = [ [ carveout.propval("reg")[1], carveout.propval("reg")[3] ] for carveout in carveouts ]

    # validate no overlaps or conflicts by generating 2d array of reg values from each reserved memory
    # this array contains reg values for such validation
    res_mem_regs = [ n.propval("reg") for n in res_mem_node.subnodes(children_only=True) if n.propval("reg") != [''] ]

    for i in range(len(res_mem_regs)):
        reg1 = res_mem_regs[i]

        # Defensive check
        if len(reg1) < addr_cells + size_cells:
            continue

        base1 = cells_to_int(reg1[:addr_cells])
        size1 = cells_to_int(reg1[addr_cells:addr_cells + size_cells])

        for j in range(i + 1, len(res_mem_regs)):
            reg2 = res_mem_regs[j]

            if len(reg2) < addr_cells + size_cells:
                continue

            base2 = cells_to_int(reg2[:addr_cells])
            size2 = cells_to_int(reg2[addr_cells:addr_cells + size_cells])
            # Only validate relevant carveouts
            if [base1, size1] not in carveout_pairs:
                continue
            # Overlap check
            if base1 < base2 + size2 and base2 < base1 + size1:
                print(
                    "ERROR: conflict between reserved memory nodes reg values:",
                    [hex(x) for x in (base1, size1, base2, size2)]
                )
                return False

    return True

def platform_validate(platform):
    """Confirm that the detected SoC platform is supported.

    Args:
        platform (SOC_TYPE): Enum representing the current platform.

    Returns:
        bool: True when the platform is one of the supported SOC_TYPE values.

    Algorithm:
        Compares the provided enum against a whitelist and prints an error when the
        platform is not supported.
    """
    if platform not in [ SOC_TYPE.ZYNQMP, SOC_TYPE.VERSAL, SOC_TYPE.VERSAL_NET, SOC_TYPE.VERSAL2 ]:
        print("ERROR: unsupported platform: ", platform)
        return False
    return True

def xlnx_remoteproc_v2_add_cluster(tree, platform, cpu_config, cluster_ranges_val, cluster_node_path):
    """Create or update the remoteproc cluster node for an RPU complex.

    Args:
        tree (LopperTree): Device tree to mutate.
        platform (SOC_TYPE): Detected SoC platform.
        cpu_config (CPU_CONFIG): RPU configuration (split or lockstep).
        cluster_ranges_val (list[int]): Flattened ``ranges`` property values.
        cluster_node_path (str): Path of the cluster node in the tree.

    Returns:
        bool: True when the cluster node is valid or successfully created.

    Algorithm:
        Derives compatibility strings/modes from the platform, verifies existing
        cluster nodes for mode consistency, merges ranges when in split mode, or
        constructs a new node populated with all required properties.
    """
    driver_compat_str  = {
      SOC_TYPE.ZYNQMP : "xlnx,zynqmp-r5fss",
      SOC_TYPE.VERSAL : "xlnx,versal-r5fss",
      SOC_TYPE.VERSAL_NET : "xlnx,versal-net-r52fss",
      SOC_TYPE.VERSAL2 : "xlnx,versal-net-r52fss",
    }

    cluster_modes = {
        CPU_CONFIG.RPU_SPLIT: 0,
        CPU_CONFIG.RPU_LOCKSTEP: 1,
    }

    cluster_node_props = {
      "compatible" : driver_compat_str[platform],
      "#address-cells": 0x2,
      "#size-cells": 0x2,
      "xlnx,cluster-mode": cluster_modes[cpu_config.value],
      "ranges": cluster_ranges_val,
    }

    # R5 cores also need tcm mode
    if platform in [ SOC_TYPE.ZYNQMP, SOC_TYPE.VERSAL ]:
        cluster_node_props["xlnx,tcm-mode"] = cluster_modes[cpu_config.value]

    try:
        cluster_node = tree[cluster_node_path]

        # here validate if cpu config does not match
        if cluster_modes[cpu_config.value] != cluster_node.propval("xlnx,cluster-mode"):
            print("ERROR: ", "split" if cpu_config == CPU_CONFIG.RPU_SPLIT else "lockstep", "for cpu value mismatches the cluster.")
            return False

        # only in split case, let range value be extended by both cores
        if cpu_config == CPU_CONFIG.RPU_SPLIT:
            cluster_node_props["ranges"].extend(cluster_ranges_val)

    except KeyError:
        cluster_node = LopperNode(-1, cluster_node_path)
        for key in cluster_node_props.keys():
            cluster_node + LopperProp(name=key, value = cluster_node_props[key])

        tree.add(cluster_node)

    return True

def xlnx_remoteproc_v2_add_core(tree, openamp_channel_info, power_domains, core_reg_val, core_reg_names, cluster_node_path, platform):
    """Insert a remoteproc core node beneath the cluster node.

    Args:
        tree (LopperTree): Device tree being updated.
        openamp_channel_info (dict): Aggregated metadata for the current channel.
        power_domains (list[int]): Flattened list of power-domain phandles/indices.
        core_reg_val (list[int]): Flattened ``reg`` values for core memories.
        core_reg_names (list[str]): Names corresponding to each ``reg`` entry.
        cluster_node_path (str): Absolute path to the cluster node.
        platform (SOC_TYPE): Detected SoC platform.

    Returns:
        LopperNode: The newly created core node.

    Algorithm:
        Determines the core node naming scheme from the platform and core index,
        builds the property set (compatibility, power domains, register ranges,
        memory regions), and attaches the node to the tree.
    """
    print(" --> xlnx_remoteproc_v2_add_core")
    compatible_strs = { SOC_TYPE.VERSAL2:  "xlnx,versal2-r52f", SOC_TYPE.VERSAL_NET:  "xlnx,versal-net-r52f", SOC_TYPE.VERSAL: "xlnx,versal-r5f", SOC_TYPE.ZYNQMP: "xlnx,zynqmp-r5f" }
    core_names = { SOC_TYPE.VERSAL_NET: "r52f", SOC_TYPE.VERSAL: "r5f", SOC_TYPE.ZYNQMP: "r5f" }
    core_names[SOC_TYPE.VERSAL2] = core_names[SOC_TYPE.VERSAL_NET]

    core_node = LopperNode(-1, "{}/{}@{}".format( cluster_node_path, core_names[platform], int(openamp_channel_info["rpu_core"])))

    core_node_props = {
      "compatible" : compatible_strs[platform],
      "power-domains": power_domains,
      "reg": core_reg_val,
      "reg-names": core_reg_names,
      "memory-region": [ n.phandle for n in openamp_channel_info["new_ddr_nodes"] ]
    }

    if openamp_channel_info["new_ddr_nodes"] == []:
        core_node_props.pop("memory-region")

    for key in core_node_props.keys():
        core_node + LopperProp(name=key, value = core_node_props[key])

    tree.add(core_node)

    return core_node


def xlnx_remoteproc_v2_cluster_base_str(platform, rpu_core):
    """Return the remoteproc cluster base address string for an RPU core.

    Args:
        platform (SOC_TYPE): Detected platform.
        rpu_core (RPU_CORE): Enum indicating the RPU core index.

    Returns:
        str: Hexadecimal string representing the base address for the cluster.

    Algorithm:
        Performs a table lookup keyed by platform and core index to derive the base
        address required for the cluster node path.
    """
    base_addresses = {
        SOC_TYPE.VERSAL_NET: {
            RPU_CORE.RPU_0: "eba00000",
            RPU_CORE.RPU_1: "eba00000",
            RPU_CORE.RPU_2: "eba40000",
            RPU_CORE.RPU_3: "eba40000",
        },
        SOC_TYPE.VERSAL2: {
            RPU_CORE.RPU_0: "eba00000",
            RPU_CORE.RPU_1: "eba00000",
            RPU_CORE.RPU_2: "ebb00000",
            RPU_CORE.RPU_3: "ebb00000",
            RPU_CORE.RPU_4: "ebc00000",
            RPU_CORE.RPU_5: "ebc00000",
            RPU_CORE.RPU_6: "ebac0000",
            RPU_CORE.RPU_7: "ebac0000",
            RPU_CORE.RPU_8: "ebbc0000",
            RPU_CORE.RPU_9: "ebbc0000",
        },
        SOC_TYPE.ZYNQMP: {
            RPU_CORE.RPU_0: "ffe00000",
            RPU_CORE.RPU_1: "ffe00000",
        },
        SOC_TYPE.VERSAL: {
            RPU_CORE.RPU_0: "ffe00000",
            RPU_CORE.RPU_1: "ffe00000",
        },
    }

    return base_addresses[platform][rpu_core]

def xlnx_remoteproc_v2_construct_cluster(tree, openamp_channel_info, channel_elfload_nodes, verbose = 0):
    """Build the remoteproc cluster and core nodes for a channel.

    Args:
        tree (LopperTree): Device tree to mutate.
        openamp_channel_info (dict): Aggregated metadata describing the channel.
        channel_elfload_nodes (list[LopperNode]): ELFLOAD carveouts referenced by the channel.
        verbose (int): Verbosity flag for diagnostic output.

    Returns:
        LopperNode | bool: Newly created core node on success, or False on failure.

    Algorithm:
        Validates platform support, merges power-domain data from carveouts, derives
        ranges for TCM and DDR nodes, ensures the cluster node exists with correct
        configuration, tracks newly added DDR regions, and finally inserts the core
        node using ``xlnx_remoteproc_v2_add_core``.
    """
    print(" -> xlnx_remoteproc_v2_construct_cluster")
 
    cpu_config = openamp_channel_info["cpu_config"]

    rpu_core = determinte_rpu_core(tree, cpu_config, openamp_channel_info["remote_node"] )
    platform = get_platform(tree, verbose)
    cluster_ranges_val = []
    core_reg_names = []
    power_domains = []
    core_reg_val = []

    global memory_nodes

    if not platform_validate(platform):
        return False

    power_domains = openamp_channel_info["rpu_core_pd_prop"].value

    r5_core_reg_names_mappings = { "ffe00000" : "atcm0", "ffe20000" : "btcm0",
                                "ffe90000" : "atcm1", "ffeb0000" : "btcm1" }
    r52_core_reg_names_mappings = { "atcm" : "atcm0", "btcm" : "btcm0", "ctcm": "ctcm0" }

    core_reg_names_mappings = r52_core_reg_names_mappings if get_platform(tree, 0) in [ SOC_TYPE.VERSAL2, SOC_TYPE.VERSAL_NET ] else r5_core_reg_names_mappings

    # loop through TCM nodes
    for n in [ n for n in channel_elfload_nodes if n.propval("xlnx,ip-name") != [''] ]:
        pd = n.propval("power-domains")
        power_domains.extend(pd)
        core_reg_val.extend(memory_nodes[pd[1]]["rpu_view"])
        cluster_ranges_val.extend(memory_nodes[pd[1]]["system_view"])

        # map TCM node name to binding compliant TCM name
        if not any(tcm_name_substr in n.name.lower() for tcm_name_substr in core_reg_names_mappings.keys()):
            print(f"ERROR: Unable to map %s to proper TCM spec name" % n.name)
            return False

        for tcm_substr in core_reg_names_mappings.keys():
            if tcm_substr in n.name:
                core_reg_names.append(core_reg_names_mappings[tcm_substr])

    # construct remoteproc cluster node
    cluster_node_path = "/remoteproc@" + xlnx_remoteproc_v2_cluster_base_str(platform, rpu_core)
    if not xlnx_remoteproc_v2_add_cluster(tree, platform, cpu_config, cluster_ranges_val, cluster_node_path):
        return False

    openamp_channel_info["new_ddr_nodes"] = [ n for n in channel_elfload_nodes if n.propval("xlnx,ip-name") == [''] ]

    # add individual core node in cluster node
    return xlnx_remoteproc_v2_add_core(tree, openamp_channel_info, power_domains,
                                       core_reg_val, core_reg_names, cluster_node_path, platform)

def xlnx_remoteproc_rpu_parse(tree, node, openamp_channel_info, elfload_nodes, verbose = 0):
    """Populate RPU-specific metadata for a remoteproc relation.

    Args:
        tree (LopperTree): Device tree being analyzed.
        node (LopperNode): Remoteproc relation child node.
        openamp_channel_info (dict): Mutable accumulator for channel information.
        elfload_nodes (list[LopperNode]): Carveout nodes used for ELF loading.
        verbose (int): Verbosity flag for diagnostic output.

    Returns:
        bool: True when parsing succeeds, False on validation errors.

    Algorithm:
        Determines CPU configuration mode, resolves the targeted RPU core index,
        validates required power-domain properties, and stores the derived values in
        ``openamp_channel_info`` for downstream processing.
    """
    print(" -> xlnx_remoteproc_rpu_parse", node)

    remote_node = openamp_channel_info["remote_node"] 
    cpu_config = determine_cpus_config(remote_node)
    if cpu_config not in [ CPU_CONFIG.RPU_LOCKSTEP, CPU_CONFIG.RPU_SPLIT]:
        print("ERROR: cpu_config: ", cpu_config, " is not in ", [ CPU_CONFIG.RPU_LOCKSTEP, CPU_CONFIG.RPU_SPLIT])
        return False

    rpu_core = determinte_rpu_core(tree, cpu_config, remote_node )
    if rpu_core not in RPU_CORE:
        print("ERROR: Invalid rpu core: ", rpu_core)
        return False

    if remote_node.propval("rpu_pd_val") == ['']:
        print("ERROR: no RPU Power domain value found")
        return False

    openamp_channel_info["rpu_core_pd_prop"] = remote_node.props("rpu_pd_val")[0]
    openamp_channel_info["cpu_config"] = cpu_config
    openamp_channel_info["rpu_core"] = str(int(rpu_core))

    return True

banner_printed = False
def get_platform(tree, verbose = 0):
    """Derive the platform enum from the root node's model/compatible strings.

    Args:
        tree (LopperTree): Device tree object.
        verbose (int): Verbosity flag controlling banner output.

    Returns:
        SOC_TYPE | None: Enum value representing the platform, or None when unknown.

    Algorithm:
        Combines the root node's ``compatible`` and ``model`` properties, optionally
        emits a banner once per execution, and scans for known substrings to map the
        tree onto a ``SOC_TYPE`` enum value.
    """
    # set platform
    global banner_printed
    platform = None
    root_node = tree["/"]

    inputs = root_node.propval("compatible") + root_node.propval("model")

    zynqmp = [ 'Xilinx ZynqMP',  "xlnx,zynqmp" ]
    versal = [ 'xlnx,versal', 'Xilinx Versal']
    versalnet = [ 'versal-net', 'Versal NET', "xlnx,versal-net", "Xilinx Versal NET" ]
    versal2 = [ 'xlnx,versal2', 'amd,versal2', 'amd versal vek385 reva' ]

    rpu_socs = [ versal2, zynqmp, versal, versalnet ]
    rpu_socs_enums = [ SOC_TYPE.VERSAL2, SOC_TYPE.ZYNQMP, SOC_TYPE.VERSAL, SOC_TYPE.VERSAL_NET ]

    if verbose > 0 and not banner_printed:
        print("[INFO]: ------> OPENAMP: XLNX: \t platform info: ", inputs)
        banner_printed = True

    for index, soc in enumerate(rpu_socs):
        for soc_str in soc:
            for i in inputs:
                if i == soc_str:
                    return rpu_socs_enums[index]

    if platform == None:
        print("Unable to find data for platform: ", root_model, root_compat)

    return platform

def openamp_nontree_outputs_handler(sdt, output_file_name, openamp_args, verbose = 0 ):
    """Derive the platform enum from the root node's model/compatible strings.
       This handler is called where outputs can be derived from the existing tree.
       typically just YAML -> DTS translation
       currently handles:
            1. BM / freertos RPU openamp header
            2. libmetal ipc .cmake output file
    Args:
        sdt (LopperSDT): Lopper system device tree with tree object stored.
        output_file_name (str): output file name
        openamp_args (Dict): dictionary of relevant arguments.
        verbose (int): Verbosity flag controlling banner output.

    Returns:
        True or False.

    Algorithm:
        Gather relation's ipi node and carveouts. Then determine the use case. Based on this
        call the output-file routine. That output-file routine shall return True or False.
    """
    print(" --> openamp_nontree_outputs_handler")
    platform = get_platform(sdt.tree, verbose)
    if platform == None:
        return False

    # get_cpu_node expects dictionary where first arg first element is machine
    machine = openamp_args['machine']

    # OS is used to determine if (a) machine should be used and (b) map to domain
    os = openamp_args["dt_type"]

    match_cpunode = get_cpu_node(sdt, {'args':[machine]}) if os != "linux_dt" else None
    if not match_cpunode and os != "linux_dt":
        print("openamp_nontree_outputs_handler: unable to find machine: ", machine)
        return False

    domains = sdt.tree['/domains']
    relation_node = None
    relation_parent_search = True if openamp_args['relation_parent'] != None else False
    compatible_string_search = True if openamp_args['compatible_string'] != None else False
    for n in domains.subnodes():
        if n.parent == None or n.parent.parent == None:
            continue

        if n.parent.parent.propval("cpus") == ['']:
            continue

        # ensure target domain matches
        if os != "linux_dt" and match_cpunode.parent != sdt.tree.pnode(n.parent.parent.propval("cpus")[0]):
            continue

        # search based on compatible string of relation
        if compatible_string_search and n.propval("compatible") != [openamp_args['compatible_string']]:
            continue

        # filter based on name
        if relation_parent_search and openamp_args['relation_parent'] != n.name:
            continue

        relation_node = n
        break

    carveouts = None
    ipi_node = None
    relation_node_search = True if openamp_args['relation'] != None else False
    for node in relation_node.subnodes(children_only=True):
        if relation_node_search and openamp_args['relation'] != node.name:
            continue

        pname = "remote" if os == "linux_dt" else "host"
        # check for remote property
        if node.props(pname) == []:
            print("ERROR: ", node, "is missing ", pname, " property")
            return False

        # first find host to remote IPI
        mbox_pval = node.propval("mbox")
        if mbox_pval == ['']:
            print("ERROR: ", node, " is missing mbox property")
            return False

        ipi_node = sdt.tree.pnode(mbox_pval[0])
        if ipi_node == None:
            print("ERROR: Unable to find ipi")
            return False

        carveout_prop = node.propval("carveouts")
        if carveout_prop == ['']:
            print("ERROR: ", node, " is missing carveouts property")
            return False

        carveouts = [ sdt.tree.pnode(phandle) for phandle in carveout_prop ]

        if not openamp_args['libmetal_output_file']:
            return xlnx_openamp_gen_outputs_only(sdt.tree, machine, output_file_name, carveouts, ipi_node, verbose)

        if [openamp_args['compatible_string']] == relation_node.propval("compatible"):
            timer_pval = node.propval("timer")
            if timer_pval == ['']:
                print("ERROR: ", node, " is missing timer property")
                return False

            timer_node = sdt.tree.pnode(node.propval("timer")[0])

            return xlnx_libmetal_gen_output_file(sdt.tree, output_file_name, carveouts, ipi_node, timer_node, os, verbose)

    return False

def xlnx_remoteproc_parse(tree, remoteproc_relation_node, carveout_validation_arr, verbose = 0 ):
    """Parse remoteproc relations and construct core nodes.

    Args:
        tree (LopperTree): Device tree being updated.
        remoteproc_relation_node (LopperNode): Domain relation describing remoteproc channels.
        carveout_validation_arr (list[LopperNode]): Accumulator for carveout validation.
        verbose (int): Verbosity level for diagnostic printing.

    Returns:
        dict[str, LopperNode] | bool: Mapping of remote node names to created core nodes,
        or False when parsing fails.

    Algorithm:
        Verifies platform support, iterates relation children, validates required
        properties, tracks ELFLOAD carveouts, enriches channel metadata via
        ``xlnx_remoteproc_rpu_parse``, constructs cluster/core nodes, and records the
        core nodes for later RPMsg processing.
    """
    print(" -> xlnx_remoteproc_parse", remoteproc_relation_node)

    # Xilinx OpenAMP subroutine to collect Remoteproc information from relation node in tree
    if get_platform(tree, verbose) == None:
        print("Unsupported platform")
        return False

    channel_to_core_dict = {}

    for node in remoteproc_relation_node.subnodes(children_only=True):
        # check for remote property
        if node.propval("remote") == ['']:
            print("ERROR: ", node, "is missing remote property")
            return False

        remote_node = tree.pnode(node.propval("remote")[0])
        openamp_channel_info = { "remote_node": remote_node }

        # check for elfload prop
        if node.props("elfload") == []:
            print("ERROR: ", node, " is missing elfload property")
            return False

        channel_elfload_nodes = [ tree.pnode(current_elfload) for current_elfload in node.propval("elfload") ]
        # validate later
        carveout_validation_arr.extend(channel_elfload_nodes)

        if not xlnx_remoteproc_rpu_parse(tree, node, openamp_channel_info, channel_elfload_nodes, verbose):
            return False

        core_node = xlnx_remoteproc_v2_construct_cluster(tree, openamp_channel_info, channel_elfload_nodes, verbose = 0)
        if not core_node:
            return False

        channel_to_core_dict[remote_node.name] = core_node # save core node for later use by rpmsg processing

    return channel_to_core_dict

def xlnx_openamp_find_compat_domains(tree, delete_nodes = False):
    """Locate or remove OpenAMP-compatible domain nodes.

    Args:
        tree (LopperTree): Device tree object.
        delete_nodes (bool): When True, delete matching domains instead of reporting them.

    Returns:
        bool: True when compatible domains are found (or removed), False otherwise.

    Algorithm:
        Scans the ``/domains`` node for children whose compatibility matches the known
        OpenAMP strings and optionally deletes them from the tree.
    """
    try:
        domain_node = tree["/domains"]
    except:
        return False

    for n in tree["/domains"].subnodes(children_only=True):
        node_compat = n.propval("compatible")
        if node_compat == ['']:
            continue
        if node_compat[0] in [ REMOTEPROC_D_TO_D_v2, RPMSG_D_TO_D, LIBMETAL_D_TO_D ]:
            if delete_nodes:
                tree - n
            else:
                return True

    return False

def parse_openamp_args(arg_inputs):
    """Parse command-line style arguments for the OpenAMP assist.

    Args:
        arg_inputs (list[str]): Argument list passed to the assist.

    Returns:
        dict: Normalized configuration containing output filename, machine name,
        and detected device tree type.

    Algorithm:
        Filters the argument list for OpenAMP-specific flags, uses argparse to obtain
        structured values, infers the device tree type from positional inputs, and
        normalizes the result for downstream consumption.
    """
    parser = argparse.ArgumentParser(description="OpenAMP argument parser")
    parser.add_argument("--openamp_output_filename", type=str, help="Output header file name")
    parser.add_argument("--openamp_remote", type=str, help="OpenAMP remote machine name")
    parser.add_argument("--openamp_header_only", action='store_true', help="OpenAMP flag to denote to only generate RPU app header. Only relevant for FreeRTOS / BM cases.")

    # This can be used for host or remote. Which means --openamp_remote is equivalent to --processor for remote case
    parser.add_argument("--processor", type=str, help="OpenAMP target processor machine name")

    parser.add_argument("--libmetal_output_file", action='store_true', help="If present - then attempt to decipher relevant IPI for the specified OpenAMP or Libmetal relation. This will also require --compatible-string and --processor. Optionally --relation-parent and --relation are used to specify non-default (e.g. first found) relation.")
    parser.add_argument("--compatible-string", type=str, help="compatible string for relation. expecting either \"libmetal,ipc-v1\" or \"openamp,rpmsg-v1\"")
    parser.add_argument("--os", type=str, help="OS arg")
    parser.add_argument("--relation-parent", type=str, help="parent of relation")
    parser.add_argument("--relation", type=str, help="target relation")

    config = {}
    if len(arg_inputs) == 2 and arg_inputs[1] in ["linux_dt", "zephyr_dt"]:
        config["dt_type"] = arg_inputs[1]
        config["machine"] = arg_inputs[0]
        for i in ["processor", "os", "libmetal_output_file", "openamp_remote", "openamp_output_filename"]:
            config[i] = None
    elif len(arg_inputs) == 1:
        config["dt_type"] = "baremetal_dt"
        config["machine"] = arg_inputs[0]
        for i in ["processor", "os", "ipi_mapping", "openamp_remote", "openamp_output_filename"]:
            config[i] = None
    else:
        args = parser.parse_args(arg_inputs)
        config = vars(args)
        config["dt_type"] = config["os"]
        config["machine"] = False

        if config["processor"] and not config["machine"]:
            config["machine"] = config["processor"]
        elif config["openamp_remote"] and config["openamp_header_only"] and not config["machine"]:
            config["machine"] = config["openamp_remote"]
        elif not config["machine"]:
            print("INFO: OpenAMP plugin: missing processor or openamp_remote being passed in. exiting now")
            return False

        # handling for ipi mapping workflow
        if config["libmetal_output_file"] and not config["compatible_string"]:
            print("requires compatible_string to be set for libmetal_output_file case")
            return False

        # provide default output file for IPI mapping use case if none provided
        if config["libmetal_output_file"] and not config["openamp_output_filename"]:
            print("INFO: OpenAMP plugin: libmetal_output_file route is taken. output file is not specified so default is used (libmetal_output_file.cmake)")
            config["openamp_output_filename"] = "libmetal_output_file.cmake"

    return config

def xlnx_openamp_parse(sdt, options, verbose = 0 ):
    """Entry point for the OpenAMP assist to process remoteproc/RPMsg data.

    Args:
        sdt (LopperSDT): Structured device tree wrapper.
        options (dict): Plugin options containing ``args``.
        verbose (int): Verbosity level for diagnostic output.

    Returns:
        bool: True when processing succeeds or no domains exist, False on errors.

    Algorithm:
        Parses assist arguments, checks for OpenAMP-compatible domains, delegates
        relation handling when appropriate.
    """
    # Xilinx OpenAMP subroutine to parse OpenAMP Channel
    # information and generate Device Tree information.
    print(" -> xlnx_openamp_parse")
    openamp_args = parse_openamp_args(options['args'])
    if not openamp_args:
        return False

    tree = sdt.tree
    ret = -1
    machine = openamp_args["machine"]

    if not xlnx_openamp_find_compat_domains(tree):
        if verbose > 1:
            print("OPENAMP: XLNX: WARNING: no openamp domains found")
        return True

    if openamp_args["openamp_output_filename"]:
        return openamp_nontree_outputs_handler(sdt, openamp_args["openamp_output_filename"], openamp_args, 1 )

    xlnx_openamp_trim_timers(sdt, openamp_args["dt_type"], machine)

    if openamp_args["dt_type"] in ["zephyr_dt", "linux_dt"] or openamp_args["openamp_output_filename"]:
        # if find_only is False, then processing will also occur.
        if not xlnx_handle_relations(sdt, machine, False, openamp_args["dt_type"]):
            return False

    return True




#/*
# * Copyright (c) 2020 Xilinx Inc. All rights reserved.
# *
# * Author:
# *       Appana Durga Kedareswara rao <appana.durga.rao@xilinx.com>
# *
# * SPDX-License-Identifier: BSD-3-Clause
# */
import sys
import types
import os
import re
from pathlib import Path
from pathlib import PurePath
from io import StringIO
import contextlib
import importlib
from lopper import Lopper
from lopper import LopperFmt
import lopper
from lopper.tree import *
from re import *
import yaml
import json

sys.path.append(os.path.dirname(__file__))
from baremetalconfig_xlnx import *
from baremetal_xparameters_xlnx import *

def is_compat(node, compat_string_to_test):
    if re.search( "module,petalinuxconfig_xlnx", compat_string_to_test):
        return xlnx_generate_petalinux_config
    return ""

class YamlDumper(yaml.Dumper):
    def increase_indent(self, flow=False, indentless=False):
        return super(YamlDumper, self).increase_indent(flow, False)

# tgt_node: is the baremetal config top level domain node number
# sdt: is the system device-tree
def xlnx_generate_petalinux_config(tgt_node, sdt, options):
    root_node = sdt.tree[tgt_node]
    root_sub_nodes = root_node.subnodes()
    compatible_list = []
    driver_list = []
    node_list = []
    symbol_node = ""
    root_compat = tgt_node.propval('compatible')
    yaml_file = options['args'][1]

    # Traverse the tree and find the nodes having status=ok property
    # and create a compatible_list from these nodes.
    for node in root_sub_nodes:
        try:
            if node.name == "__symbols__":
                symbol_node = node
            status = node["status"].value
            if "okay" in status:
                node_list.append(node)
        except:
           pass

    mapped_nodelist = get_mapped_nodes(sdt, node_list, options)
    nodename_list = []
    for node in mapped_nodelist:
        nodename_list.append(get_label(sdt, symbol_node, node))

    """
    1. Create unique dictonary based on the device_type and nodes
    2. Update the dictonray key with the match node label name
    3. For Memory Device_type #special handling
    4. For uart device_type add default baud rate 
    5. ???
    """
    tmp_dict = {}
    proc_name = ""
    device_type_dict = {}
    with open(yaml_file, 'r') as stream:
        schema = yaml.safe_load(stream)
        res_list = list(schema.keys())
        tmp_node_list = []
        device = sdt.tree['/'].propval('device_id')
        if device != ['']:
            device_type_dict['device_id'] = device[0]
        for res in res_list:
            dev_type = schema[res]['device_type']
            if re.search("processor", dev_type):
                ### Processor Handling
                match_cpunode = get_cpu_node(sdt, options)
                if match_cpunode:
                    label_name = get_label(sdt, symbol_node, match_cpunode)
                    ip_name = match_cpunode.propval('xlnx,ip-name')
                    proc_name = label_name
                    ipname = {"arch": "aarch64", "ip_name":ip_name[0]}
                    device_type_dict['processor'] = {label_name:ipname}
                    device_type_dict['processor'][label_name].update({"slaves_strings": " ".join(nodename_list)})
            elif re.search("memory", dev_type):
                ### Memory Handling
                mem_ranges, _ = get_memranges(tgt_node, sdt, options)
                index = 0
                memnode_list = sdt.tree.nodes('/memory@.*')
                for mem_name,mem in mem_ranges.items():
                    try:
                        mem_node = [node for node in memnode_list if re.search(mem_name[:-2], node.propval('xlnx,ip-name', list)[0])]
                    except:
                        mem_node = None
                    if index == 0:
                        tmp_dict['slaves'] = {mem_name: "None"}
                    else:
                        tmp_dict['slaves'].update({mem_name: "None"})
                    tmp_dict['slaves'][mem_name] = {"device_type":"memory"}
                    if mem_node:
                        if mem_node[0].propval('memory_type') != ['']:
                            tmp_dict['slaves'][mem_name] = {"device_type":mem_node[0].propval('memory_type', list)[0]}
                    tmp_dict['slaves'][mem_name].update({"ip_name":mem_name[:-2]})
                    tmp_dict['slaves'][mem_name].update({"baseaddr":hex(mem[0])})
                    tmp_dict['slaves'][mem_name].update({"highaddr":hex(mem[0] + mem[1])})
                    index += 1
            else:
                for node in mapped_nodelist:
                    compatible_list = node["compatible"].value
                    driver_compatlist = compat_list(schema[res])
                    match = [compat for compat in compatible_list if compat in driver_compatlist]
                    ipname = node.propval('xlnx,ip-name')
                    if match:
                        tmp_node_list.append(node)
                        label_name = get_label(sdt, symbol_node, node)
                        try:
                            if re.search("serial", dev_type):
                                baud_rate = node.propval('xlnx,baudrate')
                                addr,size = scan_reg_size(node, node['reg'].value, 0)
                                tmp_dict['slaves'][label_name] = {"device_type":"serial"}
                                tmp_dict['slaves'][label_name].update({"ip_name":ipname[0]})
                                tmp_dict['slaves'][label_name].update({"baseaddr":hex(addr)})
                            else:
                                tmp_dict['slaves'][label_name] = {"device_type":dev_type}
                                tmp_dict['slaves'][label_name].update({"ip_name":ipname[0]})
                        except KeyError:
                            pass

        mapped_nodelist = [node for node in mapped_nodelist if node not in tmp_node_list]
        for node in mapped_nodelist:
            label_name = get_label(sdt, symbol_node, node)
            try:
                ipname = node.propval('xlnx,ip-name')
                if ipname[0]:
                    tmp_dict['slaves'][label_name] = {"ip_name":ipname[0]}
            except:
                pass
        device_type_dict['processor'][proc_name].update(tmp_dict)

        with open("petalinux_config.yaml", "w") as fd:
            fd.write(yaml.dump(device_type_dict, Dumper=YamlDumper, default_flow_style=False, sort_keys=False, indent=4, width=32768))

    return True



#/*
# * Copyright (c) 2020 Xilinx Inc. All rights reserved.
# * Copyright (c) 2024 - 2026 Advanced Micro Devices, Inc.  All rights reserved.
# *
# * Author:
# *       Appana Durga Kedareswara rao <appana.durga.kedareswara.rao@amd.com>
# *
# * SPDX-License-Identifier: BSD-3-Clause
# */

import sys
import os
import re
import glob
import logging
import common_utils as utils
import baremetalconfig_xlnx as bm_config

from baremetaldrvlist_xlnx import xlnx_generate_bm_drvlist
from baremetallinker_xlnx import get_memranges
from lopper.log import _init, _warning, _info, _error, _debug, _level

sys.path.append(os.path.dirname(__file__))

_init(__name__)

def is_compat( node, compat_string_to_test ):
    if re.search( "module,baremetal_xparameters_xlnx", compat_string_to_test):
        return xlnx_generate_xparams
    return ""

def add_multi_buf(plat,match_cpunode,data_dict,else_ignore_data=[]):
    """To add multi buf data to the plat
    Args:
        plat(obj) : object of a plat
        match_cpunode(obj) : object of a cpu node
        data_dict(dict) : data to be added to buf
        else_ignore_data(list) : ignore the else part
    Returns:
        None
    """
    for key,value in data_dict.items():
        if match_cpunode.propval(key) != ['']:
            data = match_cpunode.propval(key, list)[0]
            plat.buf(f'\n#define {value} {data}\n')
        else:
            if key not in else_ignore_data:
                _debug(f'Property "{key}" not found in CPU node. Defaulting {value} to 0.')
                plat.buf(f'\n#define {value} 0\n')

def xlnx_generate_xparams(tgt_node, sdt, options):
    _level(utils.log_setup(options), __name__)
    root_node = sdt.tree[tgt_node]
    root_sub_nodes = root_node.subnodes()
    if options.get('outdir', {}):
        sdt.outdir = options['outdir']

    node_dict = {}
    node_list = []
    # Traverse the tree and find the nodes having status=ok property
    symbol_node = ""
    chosen_node = ""
    for node in root_sub_nodes:
        try:
            if node.name == "chosen":
                chosen_node = node
            if node.name == "__symbols__":
                symbol_node = node
            status = node["status"].value
            if "okay" in status:
                node_list.append(node)
                compat = node.propval('compatible')
                node_dict.update({node.abs_path:compat})
        except:
           pass

    repo_path_data = options['args'][1]
    is_fpd_coherent = None
    is_lpd_coherent = None
    is_pl_coherent = None
    fpd_master_list = ["psu_sata","psu_pcie"]
    lpd_master_list = ["psu_adma","psu_qspi","psu_nand","psu_sd","psu_ethernet","psu_cortexr5","psu_usb"]
    for node in node_list:
        try:
            prop_val = node['dma-coherent'].value
            if '' in prop_val:
                cci_en = True
        except:
            cci_en = None
        if cci_en:
            ip_name = node.propval('xlnx,ip-name', list)[0]
            if ip_name in fpd_master_list:
                is_fpd_coherent = True
            if ip_name in lpd_master_list:
                is_lpd_coherent = True
            if re.search("amba_pl", node.parent.name):
                is_pl_coherent = True
    drvlist = xlnx_generate_bm_drvlist(tgt_node, sdt, options)
    xparams = os.path.join(sdt.outdir, f"xparameters.h")
    plat = bm_config.DtbtoCStruct(xparams)
    plat.buf('#ifndef XPARAMETERS_H   /* prevent circular inclusions */\n')
    plat.buf('#define XPARAMETERS_H   /* by using protection macros */\n')
    total_nodes = node_list
    for drv in drvlist:
        if utils.is_file(repo_path_data):
            repo_schema = utils.load_yaml(repo_path_data)
            drv_data = repo_schema['driver']
            drv_dir = drv_data.get(drv,{}).get('vless','')
            if not drv_dir and drv_data.get(drv,{}).get('path',''):
                drv_dir = drv_data.get(drv,{}).get('path','')[0]
        else:
            drv_dir = os.path.join(repo_path_data, "XilinxProcessorIPLib", "drivers", drv)

        if not drv_dir:
            has_drivers = [dir_name for dir_name in os.listdir(utils.get_dir_path(sdt.dts)) if "drivers" in dir_name]
            if has_drivers:
                has_drivers = os.path.join(utils.get_dir_path(sdt.dts), "drivers")
                yaml_list = glob.glob(has_drivers + '/**/data/*.yaml', recursive=True)
                yaml_file_abs = [yaml for yaml in yaml_list if f"{drv}.yaml" == utils.get_base_name(yaml)]
                if yaml_file_abs:
                    yaml_file_abs = yaml_file_abs[0]
        else:
            yaml_file_abs = os.path.join(drv_dir, "data", f"{drv}.yaml")

        if utils.is_file(yaml_file_abs):
            schema = utils.load_yaml(yaml_file_abs)
            driver_compatlist = bm_config.compat_list(schema)
            driver_proplist = schema.get('required',{})
            if schema.get('additionalProperties', {}):
                if driver_proplist:
                    driver_proplist.extend(schema.get('additionalProperties',{}))
                else:
                    driver_proplist = schema.get('additionalProperties',{})
            match_nodes = []
            for node,compatible_list in node_dict.items():
                for comp in driver_compatlist:
                   match = [x for x in compatible_list if comp == x]
                   if match:
                       node1 = [x for x in node_list if (x.abs_path == node)]
                       node_list = [x for x in node_list if not(x.abs_path == node)]
                       if node1:
                           match_nodes.append(node1[0])
            if sdt.tree[tgt_node].propval('pruned-sdt') == ['']:
                match_nodes = bm_config.get_mapped_nodes(sdt, match_nodes, options)
            canonical_name = f"X{drv.upper()}"
            if schema.get('xparam_prefix',""):
                canonical_name = schema["xparam_prefix"].upper()

            for index, node in enumerate(match_nodes):
                label_name = bm_config.get_label(sdt, symbol_node, node)
                label_name = label_name.upper()
                canondef_dict = {}
                if index == 0:
                    plat.buf(f'\n#define XPAR_{canonical_name}_NUM_INSTANCES {len(match_nodes)}\n')
                for i, prop in enumerate(driver_proplist):
                    pad = 0
                    phandle_prop = 0
                    subnode_gen = 0
                    if isinstance(prop, dict):
                        pad = list(prop.values())[0]
                        prop = list(prop.keys())[0]
                        if isinstance(pad, dict):
                            subnode_prop = prop
                            prop = list(pad.keys())[0]
                            subnode_prop_list = list(pad.values())[0]
                            if prop == "subnode_phandle":
                                subnode_gen = 1
                                pad = 0
                                phandle_prop = 0
                        if pad == "phandle":
                            phandle_prop = 1

                    if i == 0:
                        plat.buf(f'\n/* Definitions for peripheral {label_name} */')

                    if prop == "reg":
                        try:
                            val, size = bm_config.scan_reg_size(node, node[prop].value, 0)
                            plat.buf(f'\n#define XPAR_{label_name}_BASEADDR {hex(val)}')
                            plat.buf(f'\n#define XPAR_{label_name}_HIGHADDR {hex(val + size -1)}')
                            canondef_dict.update({"BASEADDR":hex(val)})
                            canondef_dict.update({"HIGHADDR":hex(val + size - 1)})
                            """
                            Generate defines for all the available baseaddresses
                            """
                            na = node.parent["#address-cells"].value[0]
                            ns = node.parent["#size-cells"].value[0]
                            cells = na + ns
                            num_of_addr = int(len(node[prop].value)/cells)
                            if num_of_addr > 1:
                                for j in range(1, num_of_addr):
                                    try:
                                        val, size = bm_config.scan_reg_size(node, node[prop].value, j)
                                        plat.buf(f'\n#define XPAR_{label_name}_BASEADDR_{j} {hex(val)}')
                                    except IndexError:
                                        pass
                        except KeyError:
                            pass
                    elif prop == "compatible":
                        plat.buf(f'\n#define XPAR_{label_name}_{prop.upper()} "{node[prop].value[0]}"')
                        canondef_dict.update({prop:f'"{node[prop].value[0]}"'})
                    elif prop == "interrupts":
                        try:
                            intr = bm_config.get_interrupt_prop(sdt, node, node[prop].value)
                            plat.buf(f'\n#define XPAR_{label_name}_{prop.upper()} {intr[0]}')
                            canondef_dict.update({prop:intr[0]})
                        except KeyError:
                            _debug(f"Get interrupt prop is failed {node.name}, adding default value as {hex(0xFFFF)}")
                            intr = [0xFFFF]

                        """
                        Generate interrupt defines for all the interrupts
                        """
                        try:
                            inp =  node['interrupt-parent'].value[0]
                            intr_parent = [node for node in sdt.tree['/'].subnodes() if node.phandle == inp]
                            inc = intr_parent[0]["#interrupt-cells"].value[0]
                            num_of_intrs = int(len(node[prop].value)/inc)
                        except KeyError:
                            _debug(f"Get interrupt parent is failed {node.name}, adding default value as {0}")
                            num_of_intrs = 0
                        if num_of_intrs > 1:
                            for j in range(1, num_of_intrs):
                                try:
                                    plat.buf(f'\n#define XPAR_{label_name}_{prop.upper()}_{j} {intr[j]}')
                                except IndexError:
                                    pass

                        try:
                            intr_id = bm_config.get_interrupt_id(sdt, node, node[prop].value)
                            if node.parent.name != "":
                                if node.parent.name == "amba_pl":
                                    plat.buf(f'\n#define XPAR_FABRIC_{label_name}_INTR {intr_id[0]}')
                                    canondef_dict.update({"FABRIC":intr_id[0]})
                                    if num_of_intrs > 1:
                                        for j in range(1, num_of_intrs):
                                            plat.buf(f'\n#define XPAR_FABRIC_{label_name}_INTR_{j} {intr_id[j]}')
                                else:
                                    if node.propval('xlnx,ip-name') != ['']:
                                        ip_name = node.propval('xlnx,ip-name', list)[0]
                                        if drv == "ipipsu":
                                            plat.buf(f'\n#define XPAR_{label_name}_INTR {hex(intr_id[0]+32)}')
                                            canondef_dict.update({"INTR":hex(intr_id[0]+32)})
                        except KeyError:
                            _debug(f"Get interrupt id is failed {node.name}, adding default value as {[0xFFFF]}")
                            intr_id = [0xFFFF]
                            continue

                    elif prop == "interrupt-parent":
                        try:
                            intr_parent = bm_config.get_intrerrupt_parent(sdt, node[prop].value)
                            prop = prop.replace("-", "_")
                            plat.buf(f'\n#define XPAR_{label_name}_{prop.upper()} {hex(intr_parent)}')
                            canondef_dict.update({prop:hex(intr_parent)})
                        except KeyError:
                            pass
                    elif prop == "clocks":
                        try:
                            tclk_offset = bm_config.get_clock_offset(node)
                            clkprop_val = bm_config.get_clock_prop(sdt, node[prop].value, tclk_offset)
                            plat.buf(f'\n#define XPAR_{label_name}_{prop.upper()} {hex(clkprop_val)}')
                            canondef_dict.update({prop:hex(clkprop_val)})
                        except KeyError:
                            _warning(f"Clock property is missing for this node {node.name}, skipping the macro generation")
                            pass
                    elif prop == "child,required":
                        ipi_target_list = []
                        for j,child in enumerate(list(node.child_nodes.items())):
                            ipi_child_dict = {}
                            for k,p in enumerate(pad):
                                if type(p) is dict:
                                    break
                                if p == "nosub":
                                    continue
                                try:
                                    val = hex(child[1][p].value[0])
                                except KeyError:
                                    _debug(f"Get ipi child value is failed {node.name}, adding default value as {0xFFFF}")
                                    val = 0xFFFF
                                p = p.replace("-", "_")
                                p = p.replace("xlnx,", "")
                                plat.buf(f'\n#define XPAR_{label_name}_{j}_{p.upper()} {val}')
                                if drv == "ipipsu":
                                    ipi_child_dict[p.upper()]=val
                            if ipi_child_dict:
                                ipi_target_list.append(ipi_child_dict)
                    elif phandle_prop:
                        try:
                            prop_val = bm_config.get_phandle_regprop(sdt, prop, node[prop].value)
                            prop = prop.replace("-", "_")
                            plat.buf(f'\n#define XPAR_{label_name}_{prop.upper()} {hex(prop_val)}')
                            canondef_dict.update({prop:hex(prop_val)})
                        except KeyError:
                            pass
                    elif subnode_gen:
                        try:
                            phandle_value = node[subnode_prop].value[0]
                            prop_val = 1
                            subnode_prop = subnode_prop.replace("-", "_")
                            plat.buf(f'\n#define XPAR_{label_name}_{subnode_prop.upper()} {hex(prop_val)}')
                            canondef_dict.update({subnode_prop:hex(prop_val)})
                        except KeyError:
                            pass
                    elif prop == "ranges":
                        try:
                            device_type = node['device_type'].value[0]
                            if device_type == "pci":
                                device_ispci = 1
                        except KeyError:
                            _debug(f"Get device type is failed {node.name}, adding default value as {0}")
                            device_ispci = 0
                        if device_ispci:
                            prop_vallist = bm_config.get_pci_ranges(node, node[prop].value, pad)
                            i = 0
                            for j, prop_val in enumerate(prop_vallist):
                                if j % 2:
                                    plat.buf(f'\n#define XPAR_{label_name}_{prop.upper()}_HIGHADDR_{i} {prop_val}')
                                    cannon_prop = prop + str("_") + str("HIGHADDR") + str("_") + str(i)
                                    canondef_dict.update({cannon_prop:prop_val})
                                    i += 1
                                else:
                                    plat.buf(f'\n#define XPAR_{label_name}_{prop.upper()}_BASEADDR_{i} {prop_val}')
                                    cannon_prop = prop + str("_") + str("BASEADDR") + str("_") + str(i)
                                    canondef_dict.update({cannon_prop:prop_val})
                        # Add special handling for Zynq NAND
                        elif "arm,pl353-smc-r2p1" in node.propval('compatible'):
                            val = node[prop].value[2]
                            size = node[prop].value[3]
                            plat.buf(f'\n#define XPAR_{label_name}_BASEADDR {hex(val)}')
                            plat.buf(f'\n#define XPAR_{label_name}_HIGHADDR {hex(val + size -1)}')
                            canondef_dict.update({"BASEADDR":hex(val)})
                            canondef_dict.update({"HIGHADDR":hex(val + size - 1)})
                    else:
                        try:
                            prop_val = node[prop].value
                            # For boolean property if present LopperProp will return
                            # empty string convert it to baremetal config struct expected value
                            if '' in prop_val:
                                prop_val = [1]
                        except KeyError:
                            _debug(f"Get property value is failed for {prop} and node is {node.name}, adding default value as {0}")
                            prop_val = [0]
                            continue

                        if pad:
                            address_prop = ""
                            for pad_num in range(0,pad):
                                if pad_num == 0:
                                    address_prop = hex(node[prop].value[pad_num])
                                elif pad_num < len(node[prop].value):
                                    address_prop += f"{node[prop].value[pad_num]:08x}"
                            prop = prop.replace("-", "_")
                            prop = prop.replace("xlnx,", "")
                            if address_prop:
                                canondef_dict.update({prop:address_prop})
                                plat.buf(f'\n#define XPAR_{label_name}_{prop.upper()} {address_prop}')
                        else:
                            if ('/bits/' in prop_val):
                                prop_val = [int(prop_val[-1][3:-1], base=16)]

                            prop = prop.replace("-", "_")
                            prop = prop.replace("xlnx,", "")

                            if isinstance(prop_val[0], str):
                                if prop_val[0].replace('.','',1).isdigit():
                                    canondef_dict.update({prop:f'{prop_val[0]}'})
                                    plat.buf(f'\n#define XPAR_{label_name}_{prop.upper()} {prop_val[0]}')
                                else:
                                    canondef_dict.update({prop:f'"{prop_val[0]}"'})
                                    plat.buf(f'\n#define XPAR_{label_name}_{prop.upper()} "{prop_val[0]}"')
                            elif len(prop_val) > 1:
                                for k,item in enumerate(prop_val):
                                    cannon_prop = prop + str("_") + str(k)
                                    canondef_dict.update({cannon_prop:item})
                                    plat.buf(f'\n#define XPAR_{label_name}_{prop.upper()}_{k} {item}')
                            else:
                                canondef_dict.update({prop:hex(prop_val[0])})
                                plat.buf(f'\n#define XPAR_{label_name}_{prop.upper()} {hex(prop_val[0])}')

                plat.buf(f'\n\n/* Canonical definitions for peripheral {label_name} */')
                for prop,val in sorted(canondef_dict.items(), key=lambda e: e[0][0], reverse=False):
                    if prop == "FABRIC":
                        plat.buf(f'\n#define XPAR_FABRIC_{canonical_name}_{index}_INTR {val}')
                    else:
                        plat.buf(f'\n#define XPAR_{canonical_name}_{index}_{prop.upper()} {val}')

                #Generate Canonicals for IPI
                if drv=="ipipsu":
                    for target_channel, ipi_target_info in enumerate(ipi_target_list):
                        for key, value in ipi_target_info.items():
                            plat.buf(f'\n#define XPAR_{canonical_name}_{index}_CH{target_channel}_{key.upper()} {value}')
                plat.buf('\n')
                                    
    # Generate Defines for Generic Nodes
    if sdt.tree[tgt_node].propval('pruned-sdt') == ['']:
        node_list = bm_config.get_mapped_nodes(sdt, node_list, options)

    node_ip_count_dict = {}
    for node in node_list:
        try:
            label_name = bm_config.get_label(sdt, symbol_node, node)
            if label_name is not None:
                label_name = label_name.upper()
            val = bm_config.scan_reg_size(node, node['reg'].value, 0)
            plat.buf(f'\n/* Definitions for peripheral {label_name} */')
            plat.buf(f'\n#define XPAR_{label_name}_BASEADDR {hex(val[0])}\n')
            plat.buf(f'#define XPAR_{label_name}_HIGHADDR {hex(val[0] + val[1] - 1)}\n')

            node_ip_name = node['xlnx,ip-name'].value[0]
            if node_ip_name not in node_ip_count_dict.keys():
                node_ip_count_dict[node_ip_name] = 0
            else:
                node_ip_count_dict[node_ip_name] += 1

            canonical_name = node_ip_name.upper().replace("-", "_")

            # Check if canonical_name matches any label_name in node_list
            label_names_in_list = [
                bm_config.get_label(sdt, symbol_node, n).upper()
                for n in node_list
                if bm_config.get_label(sdt, symbol_node, n) is not None
            ]

            canonical_def_name = f'{canonical_name}_{node_ip_count_dict[node_ip_name]}'
            if canonical_def_name not in label_names_in_list:
                plat.buf(f'\n/* Canonical definitions for peripheral {label_name} */')
                plat.buf(f'\n#define XPAR_{canonical_name}_{node_ip_count_dict[node_ip_name]}_BASEADDR {hex(val[0])}\n')
                plat.buf(f'#define XPAR_{canonical_name}_{node_ip_count_dict[node_ip_name]}_HIGHADDR {hex(val[0] + val[1] - 1)}\n')
        except KeyError:
            pass

    for node in sdt.tree['/'].subnodes():
        ip_name = node.propval('xlnx,ip-name')
        hdmi_mode_prop = node.propval('xlnx,hdmi-mode')

        if hdmi_mode_prop != ['']:
            hdmi_mode = int(str(node.propval('xlnx,hdmi-mode', list)[0]) or "0", 0)
            if hdmi_mode > 0:
                if ip_name == ['v_hdmi_rxss1']:
                    plat.buf("\n#define XPAR_XV_HDMI_RX_FRL_ENABLE \n")
                elif ip_name == ['v_hdmi_txss1']:
                    plat.buf("\n#define XPAR_XV_HDMI_TX_FRL_ENABLE \n")

    # Define for Board
    if sdt.tree[tgt_node].propval('board') != ['']:
        board = sdt.tree[tgt_node].propval('board', list)[0]
        plat.buf(f"\n/*  BOARD definition */")
        plat.buf(f"\n#define XPS_BOARD_{board.upper()}\n")
    
    # Memory Node related defines
    options["args"].append("xparam")
    mem_ranges, _ = get_memranges(tgt_node, sdt, options)

    for key, value in sorted(mem_ranges.items(), key=lambda e: e[1][1], reverse=True):
        start,size = value[0], value[1]
        suffix = "ADDRESS"
        if "axi_noc" in key:
            suffix = "ADDR"
        plat.buf(f"\n#define XPAR_{key.upper()}_BASE{suffix} {hex(start)}")
        plat.buf(f"\n#define XPAR_{key.upper()}_HIGH{suffix} {hex(start + size - 1)}")

    #CPU parameters related defines
    match_cpunode = bm_config.get_cpu_node(sdt, options)
    if utils.is_file(repo_path_data):
        repo_schema = utils.load_yaml(repo_path_data)
        lib_data = repo_schema['os']
        lib_dir = lib_data.get("standalone",{}).get('vless','')
        if not lib_dir and lib_data.get("standalone",{}).get('path',''):
            srcdir = lib_data.get("standalone",{}).get('path','')[0]
    else:
        srcdir = os.path.join(repo_path_data, "lib", "bsp", "standalone")
    datadir = os.path.join(srcdir, "data")
    yaml_paths = glob.glob(f"{datadir}/*/*.yaml")
    if re.search("microblaze", match_cpunode.propval('compatible')[0]):
        for yamlfile in yaml_paths:
            name = os.path.basename(os.path.dirname(yamlfile))
            schema = utils.load_yaml(yamlfile)
            compatlist = bm_config.compat_list(schema)
            prop_list = schema.get('required', [])
            if schema.get('additionalProperties', {}):
                if prop_list:
                    prop_list.extend(schema.get('additionalProperties', {}))
                else:
                    prop_list = schema.get('additionalProperties', {})
            match = [compat for compat in compatlist if compat in match_cpunode.propval('compatible')]
            if match:
                plat.buf(f"\n\n/*  CPU parameters definition */\n")
                if match_cpunode.propval('xlnx,ip-name') != ['']:
                    ip_name = match_cpunode.propval('xlnx,ip-name', list)[0]
                else:
                    ip_name = name
                for prop in prop_list:
                    if match_cpunode.propval(prop) != ['']:
                        prop_val = match_cpunode.propval(prop, list)[0]
                        prop = prop.replace("-", "_")
                        prop = prop.replace("xlnx,", "")
                        plat.buf(f'#define XPAR_{ip_name.upper()}_{prop.upper()} {prop_val}\n')

    if re.search("microblaze-riscv", match_cpunode.propval('compatible')[0]):
        cpu_parameters={
        'xlnx,freq':"XPAR_CPU_CORE_CLOCK_FREQ_HZ",
        'xlnx,use-dcache':'XPAR_MICROBLAZE_RISCV_USE_DCACHE',
        'xlnx,dcache-line-len':'XPAR_MICROBLAZE_RISCV_DCACHE_LINE_LEN',
        'xlnx,dcache-byte-size':'XPAR_MICROBLAZE_RISCV_DCACHE_BYTE_SIZE',
        'xlnx,use-icache':'XPAR_MICROBLAZE_RISCV_USE_ICACHE',
        'xlnx,icache-line-len':'XPAR_MICROBLAZE_RISCV_ICACHE_LINE_LEN',
        'xlnx,icache-byte-size':'XPAR_MICROBLAZE_RISCV_ICACHE_BYTE_SIZE',
        'xlnx,use-fpu':'XPAR_MICROBLAZE_RISCV_USE_FPU',
        'xlnx,use-mmu':'XPAR_MICROBLAZE_RISCV_USE_MMU',
        'xlnx,use-sleep':'XPAR_MICROBLAZE_RISCV_USE_SLEEP',
        'xlnx,fault-tolerant':'XPAR_MICROBLAZE_RISCV_FAULT_TOLERANT',
        'xlnx,d-lmb':'XPAR_MICROBLAZE_RISCV_D_LMB',
        'xlnx,use-branch-target-cache':'XPAR_MICROBLAZE_RISCV_USE_BRANCH_TARGET_CACHE',
        'xlnx,branch-target-cache-size':'XPAR_MICROBLAZE_RISCV_BRANCH_TARGET_CACHE_SIZE',
        }
        ignore_else_part_lis = ['xlnx,freq','xlnx,dcache-byte-size','xlnx,icache-line-len',
                                'xlnx,icache-byte-size','xlnx,use-fpu',]
        add_multi_buf(plat,match_cpunode,cpu_parameters,ignore_else_part_lis)

    elif re.search("microblaze", match_cpunode.propval('compatible')[0]):
        mic_cpu = {'xlnx,freq':'XPAR_CPU_CORE_CLOCK_FREQ_HZ',
                   'xlnx,ddr-reserve-sa':'XPAR_MICROBLAZE_DDR_RESERVE_SA',
                   'xlnx,addr-size':'XPAR_MICROBLAZE_ADDR_SIZE'}
        add_multi_buf(plat,match_cpunode,mic_cpu)

    else:
        extra_cpu_param = {"xlnx,cpu-clk-freq-hz":"XPAR_CPU_CORE_CLOCK_FREQ_HZ",
                           "xlnx,timestamp-clk-freq":"XPAR_CPU_TIMESTAMP_CLK_FREQ",}
        add_multi_buf(plat,match_cpunode,extra_cpu_param)

    #PSS REF clocks define
    if match_cpunode.propval('xlnx,pss-ref-clk-freq') != ['']:
        pss_ref = match_cpunode.propval('xlnx,pss-ref-clk-freq', list)[0]
        plat.buf(f'#define XPAR_PSU_PSS_REF_CLK_FREQ_HZ {pss_ref}\n')
    #Define for NUMBER_OF_SLRS
    if sdt.tree[tgt_node].propval('slrcount') != ['']:
        val = sdt.tree[tgt_node].propval('slrcount', list)[0]
        plat.buf(f"\n/* Number of SLRs */")
        plat.buf(f"\n#define NUMBER_OF_SLRS {hex(val)}\n")

    #Define for DEVICE_ID
    if sdt.tree[tgt_node].propval('device_id') != ['']:
        val = sdt.tree[tgt_node].propval('device_id', list)[0]
        plat.buf(f"\n/* Device ID */")
        plat.buf(f'\n#define XPAR_DEVICE_ID "{val}"\n')

    #Define for DDRMC5_DEVICE_TYPE
    if sdt.tree[tgt_node].propval('xlnx,ddrmc5-device-type') != ['']:
        val = sdt.tree[tgt_node].propval('xlnx,ddrmc5-device-type', list)[0]
        plat.buf(f"\n#define NOC_MC_DDR5\n")
        plat.buf(f"\n/* DDRMC5_DEVICE_TYPE */")
        if val == "RDIMMs":
            plat.buf(f"\n#define DDRMC5_DEVICE_TYPE_RDIMM\n")
        elif val == "UDIMMs":
            plat.buf(f"\n#define DDRMC5_DEVICE_TYPE_UDIMM\n")

    #Define for DDRMC5_I2C_MASTER and DDRMC5_DEBUG_ELF
    if sdt.tree[tgt_node].propval('xlnx,ddrmc5-i2c-master') != ['']:
        val = sdt.tree[tgt_node].propval('xlnx,ddrmc5-i2c-master', list)[0]
        plat.buf(f"\n/* DDRMC5_I2C_MASTER */")
        plat.buf(f'\n#define DDRMC5_I2C_MASTER "{val}"\n')

    if sdt.tree[tgt_node].propval('xlnx,ddrmc5-debug-elf') != ['']:
        val = sdt.tree[tgt_node].propval('xlnx,ddrmc5-debug-elf', list)[0]
        plat.buf(f"\n/* DDRMC5_DEBUG_ELF */")
        plat.buf(f'\n#define DDRMC5_DEBUG_ELF "{val}"\n')

    #Define for XSEM_CFRSCAN_EN
    if sdt.tree[tgt_node].propval('semmem-scan') != ['']:
        val = sdt.tree[tgt_node].propval('semmem-scan', list)[0]
        plat.buf(f'\n#define XSEM_CFRSCAN_EN {val}\n')

    #Define for XSEM_NPISCAN_EN
    if sdt.tree[tgt_node].propval('semnpi-scan') != ['']:
        val = sdt.tree[tgt_node].propval('semnpi-scan', list)[0]
        plat.buf(f'\n#define XSEM_NPISCAN_EN {val}\n')

    if is_fpd_coherent:
        plat.buf("\n#define XPAR_FPD_IS_CACHE_COHERENT \n")
    if is_lpd_coherent:
        plat.buf("\n#define XPAR_LPD_IS_CACHE_COHERENT \n")
    if is_pl_coherent:
        plat.buf("\n#define XPAR_PL_IS_CACHE_COHERENT \n")

    plat.buf('\n#endif  /* end of protection macro */')
    plat.out(''.join(plat.get_buf()))

    return True



#/*
# * Copyright (c) 2020 Xilinx Inc. All rights reserved.
# *
# * Author:
# *       Bruce Ashfield <bruce.ashfield@xilinx.com>
# *
# * SPDX-License-Identifier: BSD-3-Clause
# */

from flask import Flask
from flask import jsonify
from flask import send_file
from flask_restful import Resource, Api, reqparse
import pandas as pd
import ast
import tempfile
import json

from collections import OrderedDict
import lopper

app = Flask(__name__)
api = Api(app)

sdt = None

class Domains(Resource):

    def get(self):
        try:
            domains = sdt.tree.nodes( "/domains/[^/]*$" )
        except:
            domains = []

        domain_names = ""
        if domains:
            # this just dumps the domain names comma separated. We loop
            # up to the last element, adding the ",". We then add the
            # last element after the loop, no "," .. we could also just
            # remove the trailing "," after the loop, but the list slicing
            # seems a bit faster.
            for d in domains[:-1]:
                domain_names += d.abs_path + ","
            domain_names = domain_names + domains[-1].abs_path

        return domain_names, 200

class Tree(Resource):
    def get(self):
        fpp = tempfile.NamedTemporaryFile( delete=True )

        if sdt:
            lp = lopper.LopperTreePrinter( sdt.tree.fdt, False, fpp.name )
            lp.exec()

            with open(fpp.name, 'r') as f:
                filecontents = f.read()

            # json = json.dumps(filecontents, ensure_ascii=False)
            return filecontents, 200
        else:
            return "", 204

class Nodes(Resource):
    def get(self):
        parser = reqparse.RequestParser()

        parser.add_argument('path', required=True)
        parser.add_argument('details', required=False)
        args = parser.parse_args()

        try:
            details = args['details']
            if details == "True":
                details = True
            else:
                details = False
        except:
            details = False

        node_data = OrderedDict()

        node_list = sdt.tree.nodes( args['path'] )

        if not details:
            for n in node_list:
                node_data[n.abs_path] = None
        else:
            for n in node_list:
                prop_dict = OrderedDict()
                for p in n.__props__:
                    prop_dict[p] = n.__props__[p].string_val

                node_data[n.abs_path] = prop_dict

        return node_data, 200

api.add_resource(Domains, '/domains')  # '/domains' is an entry point
api.add_resource(Tree, '/tree')  # '/tree' is an entry point
api.add_resource(Nodes, '/nodes')  # '/tree' is an entry point



# /*
# * Copyright (c) 2021 Xilinx Inc. All rights reserved.
# *
# * Author:
# *       Izhar Shaikh <izhar.ameer.shaikh@xilinx.com>
# *
# * SPDX-License-Identifier: BSD-3-Clause
# */

import xppu
from lopper import Lopper
import lopper
from lopper.tree import *

sys.path.append(os.path.dirname(__file__))


def write_cmd(addr, val):
    cmd = "write {0} {1}".format(addr, val)
    return cmd


def write_cmd_print(addr, val, fp=None):
    print(write_cmd(hex(addr), hex(val)), file=fp)


def mask_write_cmd(addr, mask, val):
    cmd = "mask_write {0} {1} {2}".format(addr, mask, val)
    return cmd


def pm_init_node_cmd1(pm_id, val):
    cmd = "pm_init_node {0} {1} {2}".format(pm_id, hex(0x1), val)
    return cmd


def pm_init_node_cmd2(pm_id, aper, val):
    cmd = "pm_init_node {0} {1} {2} {3}".format(pm_id, hex(0x2), hex(aper),
                                                val)
    return cmd


def write_header(x, fp=None):
    print("marker", hex(0x64), '"{}"'.format(x.name.upper()), file=fp)
    cmt = "# {0} : Setup START".format(x.name.upper())
    print(cmt, file=fp)


def write_footer(x, fp=None):
    cmt = "# {0} : Setup END".format(x.name.upper())
    print(cmt, file=fp)
    print("marker", hex(0x65), '"{}"'.format(x.name.upper()), file=fp)


def write_xppu_mid_list(x, fp=None):
    for i in range(xppu.MASTERS):
        master = x.get_master(i)
        if master is None:
            continue
        cmt = "# {0} XPPU: Set Master ID {1} to {2}".format(
            x.name, i, master.name)
        cmt += "\n# [MID: {0}, Mask: {1}, {2}, Parity: {3}]".format(
            master.mid,
            master.mask,
            "Read Only" if master.rw == 1 else "Read/Write",
            "Enable" if master.parity == 1 else "Disable",
        )
        cmd = write_cmd(x.get_master_addr(i), x.get_master_val(i))
        print(cmt, file=fp)
        print(cmd, file=fp)


def write_xppu_ctrl_reg(x, fp=None):
    cmt = "# Parity error checking for Aperture and Master ID entries\n"
    cmt += "# APER_PARITY_EN = {0}\n".format(xppu.APER_PARITY)
    cmt += "# MID_PARITY_EN = {0}".format(xppu.MID_PARITY)
    val = x.get_ctrl_reg_val()
    cmd = mask_write_cmd(x.baseaddr, val, val)
    print(cmt, file=fp)
    print(cmd, file=fp)


def write_xppu_ien_reg(x, fp=None):
    cmt = "# Interrupt Enable\n"
    for istr, ival in xppu.Interrupts.items():
        cmt += "# {0} = {1}\n".format(istr.upper(), ival[0])
    addr, val = x.get_ien_reg_addr_val()
    cmd = write_cmd(addr, val)
    print(cmt + cmd, file=fp)


def write_xppu_en_cmd(x, fp=None):
    cmt = "# Enable {0} XPPU\n".format(x.name)
    cmt += "# Initialize all apertures with default aperture value\n"
    # cmt += "# Default masters are {0}".format(xppu.DEF_MASTERS)
    def_aperture = x.get_default_aperture()
    cmt += "# Default masters are {0}".format(
        x.get_master_list_from_aperture(xppu.h2i(def_aperture)))
    cmd = pm_init_node_cmd1(x.pm_id, def_aperture)
    print(cmt, file=fp)
    print(cmd, file=fp)


# --
# NOTE:
# These are temporary hard-coded apertures required for the system to boot
# when the protection configuration is generated for subsystems.
# There is no way to express this in the SDT/YAML today, therefore it is directly
# patched in the output CDO. We'll remove this when SDT/YAML spec gaps are resolved.
# --
def write_xppu_allowed_entries(x, fp=None):
    # for PMC_XPPU
    pmc_xppu_fixes = """
        # FIXME
        # ATF: Allow PMC_GLOBAL (17) to be APU RD Only
        pm_init_node 0x24000002 0x2 0x11 0x8001fe1
        # ATF: Allow (386) to be APU RD Only
        pm_init_node 0x24000002 0x2 0x182 0x8001fe1

        # Uboot/Linux: Allow CRP (38) to be APU RD Only
        pm_init_node 0x24000002 0x2 0x26 0x8001fe1
        # Linux: Allow PMC_SYSMON (39) [Only Root] to be APU RW
        pm_init_node 0x24000002 0x2 0x27 0x80007f9
    """
    # for LPD_XPPU
    lpd_xppu_fixes = """
        # FIXME
        # ATF: Allow CRL (94) to be APU RW
        pm_init_node 0x24000001 0x2 0x5E 0x80007f9
        # ATF: Allow IOU_SCNTRS (20) to be APU RW
        pm_init_node 0x24000001 0x2 0x14 0x80007f9

        # Linux: Allow USB_XHCI (386) to be APU RW
        pm_init_node 0x24000001 0x2 0x182 0x80007f9
    """
    block = None

    if x.pm_id == "0x24000002":
        block = pmc_xppu_fixes
    elif x.pm_id == "0x24000001":
        block = lpd_xppu_fixes
    else:
        return

    if block is not None:
        for line in block.split("\n"):
            print(line.strip(), file=fp)


def write_xppu(x, fp=None):
    write_header(x, fp)
    write_xppu_mid_list(x, fp)
    write_xppu_ctrl_reg(x, fp)
    write_xppu_ien_reg(x, fp)
    write_xppu_en_cmd(x, fp)
    write_xppu_allowed_entries(x, fp)
    write_footer(x, fp)


def write_xmpu_regions(x, fp=None):
    for i, region in x.regions.items():
        if not region.config.enable:
            continue

        cmt = "# Region {0}".format(i)
        print(cmt, file=fp)

        # addr start
        addr, val = region.get_start_lo()
        write_cmd_print(x.baseaddr + addr, val, fp)
        addr, val = region.get_start_hi()
        write_cmd_print(x.baseaddr + addr, val, fp)

        # addr end
        addr, val = region.get_end_lo()
        write_cmd_print(x.baseaddr + addr, val, fp)
        addr, val = region.get_end_hi()
        write_cmd_print(x.baseaddr + addr, val, fp)

        # master
        addr, val = region.get_master()
        write_cmd_print(x.baseaddr + addr, val, fp)

        # config
        addr, val = region.get_config()
        write_cmd_print(x.baseaddr + addr, val, fp)


def write_xmpu_ien_reg(x, fp):
    en_regions = [r for _, r in x.regions.items() if r.config.enable]
    if len(en_regions) <= 0:
        return
    cmt = "# Interrupt Enable\n"
    addr, val = x.get_ien_reg_addr_val()
    cmt += write_cmd(addr, val)
    print(cmt, file=fp)


def write_xmpu(x, fp=None):
    write_header(x, fp)
    write_xmpu_regions(x, fp)
    write_xmpu_ien_reg(x, fp)
    write_footer(x, fp)



"""Xilinx OpenAMP common helpers shared across assist modules.

This module defines enumerations representing CPU configuration state, the
memory map metadata consumed during remoteproc construction, and helper
routines that resolve references within generated device-tree relations.
Global collections exposed here are documented with inline docstrings so they
can be surfaced by documentation tooling.
"""

from lopper.tree import *
from enum import Enum
from enum import IntEnum
import ast

class CPU_CONFIG(IntEnum):
    """Enumerate the supported RPU execution configurations."""

    RPU_SPLIT = 0
    RPU_LOCKSTEP = 1

class RPU_CORE(IntEnum):
    """Enumerate individual RPU core indices."""

    RPU_0 = 0
    RPU_1 = 1
    RPU_2 = 2
    RPU_3 = 3
    RPU_4 = 4
    RPU_5 = 5
    RPU_6 = 6
    RPU_7 = 7
    RPU_8 = 8
    RPU_9 = 9

# This is used for YAML representation
# after this is parsed, the above enums are used for internal record keeping.
class CLUSTER_CONFIG(Enum):
    """Enumerate cluster configuration strings used by YAML input."""

    RPU_LOCKSTEP = 0
    RPU_0 = 1
    RPU_1 = 2

memory_nodes = {
    15: {  # psu_r5_0_atcm_global
        "system_view": [0x0, 0x0, 0x0, 0xffe00000, 0x0, 0x10000],
        "rpu_view": [0x0, 0x0, 0x0, 0x10000]
    },
    16: {  # psu_r5_0_btcm_global
        "system_view": [0x0, 0x20000, 0x0, 0xffe20000, 0x0, 0x10000],
        "rpu_view": [0x0, 0x20000, 0x0, 0x10000]
    },
    17: {  # psu_r5_1_atcm_global
        "system_view": [0x1, 0x0, 0x0, 0xffe90000, 0x0, 0x10000],
        "rpu_view": [0x1, 0x0, 0x0, 0x10000]
    },
    18: {  # psu_r5_1_btcm_global
        "system_view": [0x1, 0x20000, 0x0, 0xffeb0000, 0x0, 0x10000],
        "rpu_view": [0x1, 0x20000, 0x0, 0x10000]
    },
    0x1831800b: {  # psv_r5_0_atcm_global
        "system_view": [0x0, 0x0, 0x0, 0xffe00000, 0x0, 0x10000],
        "rpu_view": [0x0, 0x0, 0x0, 0x10000]
    },
    0x1831800c: {  # psv_r5_0_btcm_global
        "system_view": [0x0, 0x20000, 0x0, 0xffe20000, 0x0, 0x10000],
        "rpu_view": [0x0, 0x20000, 0x0, 0x10000]
    },
    0x1831800d: {  # psv_r5_1_atcm_global
        "system_view": [0x1, 0x0, 0x0, 0xffe90000, 0x0, 0x10000],
        "rpu_view": [0x1, 0x0, 0x0, 0x10000]
    },
    0x1831800e: {  # psv_r5_1_btcm_global
        "system_view": [0x1, 0x20000, 0x0, 0xffeb0000, 0x0, 0x10000],
        "rpu_view": [0x1, 0x20000, 0x0, 0x10000]
    },
    0x183180cb: {  # r52_0a_atcm_global
        "system_view": [0x0, 0x0, 0x0, 0xeba00000, 0x0, 0x10000],
        "rpu_view": [0x0, 0x0, 0x0, 0x10000]
    },
    0x183180cc: {  # r52_0a_btcm_global
        "system_view": [0x0, 0x10000, 0x0, 0xeba10000, 0x0, 0x8000],
        "rpu_view": [0x0, 0x10000, 0x0, 0x8000]
    },
    0x183180cd: {  # r52_0a_ctcm_global
        "system_view": [0x0, 0x18000, 0x0, 0xeba20000, 0x0, 0x8000],
        "rpu_view": [0x0, 0x18000, 0x0, 0x8000]
    },
    0x183180ce: {  # r52_0b_atcm_global
        "system_view": [0x1, 0x0, 0x0, 0xEBA80000, 0x0, 0x10000],
        "rpu_view": [0x1, 0x0, 0x0, 0x10000]
    },  
    0x183180cf: {  # r52_0b_btcm_global
        "system_view": [0x1, 0x10000, 0x0, 0xEBA90000, 0x0, 0x8000],
        "rpu_view": [0x1, 0x10000, 0x0, 0x8000]
    },  
    0x183180d0: {  # r52_0b_ctcm_global
        "system_view": [0x1, 0x18000, 0x0, 0xEBAA0000, 0x0, 0x8000],
        "rpu_view": [0x1, 0x18000, 0x0, 0x8000]
    },
    0x18318106: {  # r52_0d_atcm_global
        "system_view": [0x0, 0x0, 0x0, 0xEBC00000, 0x0, 0x10000],
        "rpu_view": [0x0, 0x0, 0x0, 0x10000]
    },  
    0x18318107: {  # r52_0d_btcm_global
        "system_view": [0x0, 0x10000, 0x0, 0xEBC10000, 0x0, 0x8000],
        "rpu_view": [0x0, 0x10000, 0x0, 0x8000]
    },  
    0x18318108: {  # r52_0d_ctcm_global
        "system_view": [0x0, 0x18000, 0x0, 0xEBC20000, 0x0, 0x8000],
        "rpu_view": [0x0, 0x18000, 0x0, 0x8000]
    },  
    0x18318109: {  # r52_1d_atcm_global
        "system_view": [0x1, 0x0, 0x0, 0xEBC40000, 0x0, 0x10000],
        "rpu_view": [0x1, 0x0, 0x0, 0x10000]
    },
    0x1831810a: {  # r52_1d_btcm_global
        "system_view": [0x1, 0x10000, 0x0, 0xEBC50000, 0x0, 0x8000],
        "rpu_view": [0x1, 0x10000, 0x0, 0x8000]
    },
    0x1831810b: {  # r52_1d_ctcm_global
        "system_view": [0x1, 0x18000, 0x0, 0xEBC60000, 0x0, 0x8000],
        "rpu_view": [0x1, 0x18000, 0x0, 0x8000]
    },
    0x1831810c: {  # r52_0e_atcm_global
        "system_view": [0x0, 0x0, 0x0, 0xEBC00000, 0x0, 0x10000],
        "rpu_view": [0x0, 0x0, 0x0, 0x10000]
    },  
    0x1831810d: {  # r52_0e_btcm_global
        "system_view": [0x0, 0x10000, 0x0, 0xEBC10000, 0x0, 0x8000],
        "rpu_view": [0x0, 0x10000, 0x0, 0x8000]
    },  
    0x1831810e: {  # r52_0e_ctcm_global
        "system_view": [0x0, 0x18000, 0x0, 0xEBC20000, 0x0, 0x8000],
        "rpu_view": [0x0, 0x18000, 0x0, 0x8000]
    },  
    0x1831810f: {  # r52_1e_atcm_global
        "system_view": [0x1, 0x0, 0x0, 0xEBC40000, 0x0, 0x10000],
        "rpu_view": [0x1, 0x0, 0x0, 0x10000]
    },
    0x18318110: {  # r52_1e_btcm_global
        "system_view": [0x1, 0x10000, 0x0, 0xEBC50000, 0x0, 0x8000],
        "rpu_view": [0x1, 0x10000, 0x0, 0x8000]
    },
    0x18318111: {  # r52_1e_ctcm_global
        "system_view": [0x1, 0x18000, 0x0, 0xEBC60000, 0x0, 0x8000],
        "rpu_view": [0x1, 0x18000, 0x0, 0x8000]
    },
}
"""dict[int, dict[str, list[int]]]: Mapping of power-domain identifiers to
system-view and RPU-view memory descriptors used for remoteproc construction."""

openamp_linux_hosts = [ "psv_cortexa72_0", "psx_cortexa78_0", "psu_cortexa53_0", "cortexa78_0" ]
"""list[str]: Names of processor nodes recognized as OpenAMP Linux hosts."""

openamp_remotes = { "psx_cortexr52_0", "psx_cortexr52_1", "psx_cortexr52_2", "psx_cortexr52_3",
 "cortexr52_0", "cortexr52_1", "cortexr52_2", "cortexr52_3", "cortexr52_4",
 "cortexr52_5", "cortexr52_6", "cortexr52_7", "cortexr52_8", "cortexr52_9",
 "psu_cortexr5_0", "psu_cortexr5_1", "psv_cortexr5_1", "psv_cortexr5_0", }
"""set[str]: Names of processor nodes supported as OpenAMP remotes."""

class SOC_TYPE:
    """Enum-like constants for supported SoC families."""

    UNINITIALIZED = -1
    VERSAL = 0
    ZYNQMP = 1
    ZYNQ = 2
    VERSAL_NET = 3
    VERSAL2 = 4

platform_info_header_r5_template = """
/*
 * Copyright (c) 2025 AMD, Inc.
 * All rights reserved.
 *
 * SPDX-License-Identifier: BSD-3-Clause
 */

#ifndef _AMD_GENERATED_H_
#define _AMD_GENERATED_H_

/* Interrupt vectors */
#if defined (FREERTOS_BSP) || defined (USE_FREERTOS)
#define IPI_IRQ_VECT_ID         $IPI_IRQ_VECT_ID_FREERTOS
#else
#define IPI_IRQ_VECT_ID         $IPI_IRQ_VECT_ID
#endif
#define POLL_BASE_ADDR          $POLL_BASE_ADDR
#define IPI_CHN_BITMASK         $IPI_CHN_BITMASK

#define NUM_VRINGS              0x02
#define VRING_ALIGN             0x1000
#define VRING_SIZE              256

#define RING_TX                 $RING_TX
#define RING_RX                 $RING_RX

#define SHARED_MEM_PA           $SHARED_MEM_PA
#define SHARED_MEM_SIZE         $SHARED_MEM_SIZE
#define SHARED_BUF_OFFSET       $SHARED_BUF_OFFSET

#define SHM_DEV_NAME            $SHM_DEV_NAME
#define DEV_BUS_NAME            $DEV_BUS_NAME
#define IPI_DEV_NAME            $IPI_DEV_NAME

$EXTRAS

#endif /* _AMD_GENERATED_H_ */
"""
"""str: Template used to generate OpenAMP R5 platform header files."""

libmetal_cmake_template = """
# ============================================================================
# AUTOGENERATED FILE — DO NOT EDIT
# Generated by Python (string.Template)
# ============================================================================

# ------------------------------
# Shared Memory Devices
# ------------------------------
set(SHM_DEV_NAME              "$SHM_DEV_NAME")
set(SHM0_DESC_DEV_NAME        "$SHM0_DESC_DEV_NAME")
set(SHM1_DESC_DEV_NAME        "$SHM1_DESC_DEV_NAME")

# ------------------------------
# Base Addresses
# ------------------------------
set(SHM_IMAGE_BASE            $SHM_IMAGE_BASE)
set(SHM0_DESC_BASE            $SHM0_DESC_BASE)
set(SHM1_DESC_BASE            $SHM1_DESC_BASE)

# ------------------------------
# Sizes
# ------------------------------
set(SHM_IMAGE_SIZE            $SHM_IMAGE_SIZE)
set(SHM0_DESC_SIZE            $SHM0_DESC_SIZE)
set(SHM1_DESC_SIZE            $SHM1_DESC_SIZE)

# ------------------------------
# Payload Configuration
# ------------------------------
set(SHM_PAYLOAD_BASE          $SHM_PAYLOAD_BASE)
set(SHM_PAYLOAD_SIZE          $SHM_PAYLOAD_SIZE)
set(SHM_PAYLOAD_HALF_SIZE     $SHM_PAYLOAD_HALF_SIZE)

set(SHM_PAYLOAD_RX_OFFSET     $SHM_PAYLOAD_RX_OFFSET)
set(SHM_PAYLOAD_TX_OFFSET     $SHM_PAYLOAD_TX_OFFSET)

# ------------------------------
# Aggregate SHM Region
# ------------------------------
set(SHM_BASE_ADDR             $SHM_BASE_ADDR)
set(SHM_SIZE                  $SHM_SIZE)

# ------------------------------
# IPI Configuration
# ------------------------------
set(IPI_DEV_NAME              "$IPI_DEV_NAME")
set(IPI_BASE_ADDR             $IPI_BASE_ADDR)
set(IPI_MASK                  $IPI_MASK)
set(IPI_IRQ_VECT_ID           $IPI_IRQ_VECT_ID)

# ------------------------------
# TTC Configuration
# ------------------------------
set(TTC_DEV_NAME              "$TTC_DEV_NAME")
set(TTC_NODEID                $TTC_NODEID)
set(TTC_BASE_ADDR            $TTC_BASE_ADDR)

# ------------------------------
# Bus Configuration
# ------------------------------
set(BUS_NAME                  "$BUS_NAME")
"""
